import asyncio
import gc
import os

import pytest

import aiomysql


@pytest.fixture()
def fill_my_cnf(mysql_params):
    tests_root = os.path.abspath(os.path.dirname(__file__))

    if "unix_socket" in mysql_params:
        tmpl_path = "fixtures/my.cnf.unix.tmpl"
    else:
        tmpl_path = "fixtures/my.cnf.tcp.tmpl"

    path1 = os.path.join(tests_root, tmpl_path)
    path2 = os.path.join(tests_root, 'fixtures/my.cnf')
    with open(path1) as f1:
        tmpl = f1.read()
    with open(path2, 'w') as f2:
        f2.write(tmpl.format_map(mysql_params))


@pytest.mark.run_loop
async def test_connect_timeout(connection_creator):
    # All exceptions are caught and raised as operational errors
    with pytest.raises(aiomysql.OperationalError):
        await connection_creator(connect_timeout=0.000000000001)


@pytest.mark.run_loop
async def test_config_file(fill_my_cnf, connection_creator, mysql_params):
    tests_root = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(tests_root, 'fixtures/my.cnf')
    conn = await connection_creator(read_default_file=path)

    if "unix_socket" in mysql_params:
        assert conn.unix_socket == mysql_params["unix_socket"]
    else:
        assert conn.host == mysql_params['host']
        assert conn.port == mysql_params['port']
    assert conn.user, mysql_params['user']

    # make sure connection is working
    cur = await conn.cursor()
    await cur.execute('SELECT 42;')
    (r, ) = await cur.fetchone()
    assert r == 42
    conn.close()


@pytest.mark.run_loop
async def test_config_file_with_different_group(fill_my_cnf,
                                                connection_creator,
                                                mysql_params):
    # same test with config file but actual settings
    # located in not default group.
    tests_root = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(tests_root, 'fixtures/my.cnf')
    group = 'client_with_unix_socket'
    conn = await connection_creator(read_default_file=path,
                                    read_default_group=group)

    assert conn.charset == 'utf8'
    assert conn.user == 'root'

    # make sure connection is working
    cur = await conn.cursor()
    await cur.execute('SELECT 42;')
    (r, ) = await cur.fetchone()
    assert r == 42
    conn.close()


@pytest.mark.run_loop
async def test_utf8mb4(connection_creator):
    """This test requires MySQL >= 5.5"""
    charset = 'utf8mb4'
    conn = await connection_creator(charset=charset)
    assert conn.charset == charset
    conn.close()


@pytest.mark.run_loop
async def test_largedata(connection_creator):
    """Large query and response (>=16MB)"""
    conn = await connection_creator()
    cur = await conn.cursor()
    await cur.execute("SELECT @@max_allowed_packet")
    r = await cur.fetchone()
    if r[0] < 16 * 1024 * 1024 + 10:
        pytest.skip('Set max_allowed_packet to bigger than 17MB')
    else:
        t = 'a' * (16 * 1024 * 1024)
        await cur.execute("SELECT '" + t + "'")
        r = await cur.fetchone()
        assert r[0] == t


@pytest.mark.run_loop
async def test_escape_string(connection_creator):
    con = await connection_creator()
    cur = await con.cursor()

    assert con.escape("foo'bar") == "'foo\\'bar'"
    # literal is alias for escape
    assert con.literal("foo'bar") == "'foo\\'bar'"
    await cur.execute("SET sql_mode='NO_BACKSLASH_ESCAPES'")
    assert con.escape("foo'bar") == "'foo''bar'"


@pytest.mark.run_loop
async def test_sql_mode_param(connection_creator):
    con = await connection_creator(sql_mode='NO_BACKSLASH_ESCAPES')
    assert con.escape("foo'bar") == "'foo''bar'"


@pytest.mark.run_loop
async def test_init_param(connection_creator):
    init_command = "SET sql_mode='NO_BACKSLASH_ESCAPES';"
    con = await connection_creator(init_command=init_command)
    assert con.escape("foo'bar") == "'foo''bar'"


@pytest.mark.run_loop
async def test_autocommit(connection_creator):
    con = await connection_creator()
    assert con.get_autocommit() is False

    cur = await con.cursor()
    await cur.execute("SET AUTOCOMMIT=1")
    assert con.get_autocommit() is True

    await con.autocommit(False)
    assert con.get_autocommit() is False
    await cur.execute("SELECT @@AUTOCOMMIT")
    r = await cur.fetchone()
    assert r[0] == 0


@pytest.mark.run_loop
async def test_select_db(connection_creator):
    con = await connection_creator()
    current_db = 'test_pymysql'
    other_db = 'test_pymysql2'
    cur = await con.cursor()
    await cur.execute('SELECT database()')
    r = await cur.fetchone()
    assert r[0] == current_db

    await con.select_db(other_db)
    await cur.execute('SELECT database()')
    r = await cur.fetchone()
    assert r[0] == other_db


@pytest.mark.run_loop
async def test_connection_gone_away(connection_creator):
    # test
    # http://dev.mysql.com/doc/refman/5.0/en/gone-away.html
    # http://dev.mysql.com/doc/refman/5.0/en/error-messages-client.html
    # error_cr_server_gone_error
    conn = await connection_creator()
    cur = await conn.cursor()
    await cur.execute("SET wait_timeout=1")
    await asyncio.sleep(2)
    with pytest.raises(aiomysql.OperationalError) as cm:
        await cur.execute("SELECT 1+1")
    # error occures while reading, not writing because of socket buffer.
    # assert cm.exception.args[0] == 2006
    assert cm.value.args[0] in (2006, 2013)
    conn.close()


@pytest.mark.run_loop
async def test_connection_info_methods(connection_creator, mysql_params):
    conn = await connection_creator()
    # trhead id is int
    assert isinstance(conn.thread_id(), int)
    assert conn.character_set_name() in ('latin1', 'utf8mb4')
    if "unix_socket" in mysql_params:
        assert mysql_params["unix_socket"] in conn.get_host_info()
    else:
        assert str(conn.port) in conn.get_host_info()
    assert isinstance(conn.get_server_info(), str)
    # protocol id is int
    assert isinstance(conn.get_proto_info(), int)
    conn.close()


@pytest.mark.run_loop
async def test_connection_set_charset(connection_creator):
    conn = await connection_creator()
    assert conn.character_set_name(), ('latin1' in 'utf8mb4')
    await conn.set_charset('utf8')
    assert conn.character_set_name() == 'utf8'


@pytest.mark.run_loop
async def test_connection_ping(connection_creator):
    conn = await connection_creator()
    await conn.ping()
    assert conn.closed is False
    conn.close()
    await conn.ping()
    assert conn.closed is False


@pytest.mark.run_loop
async def test_connection_properties(connection_creator, mysql_params):
    conn = await connection_creator()
    if "unix_socket" in mysql_params:
        assert conn.unix_socket == mysql_params["unix_socket"]
    else:
        assert conn.host == mysql_params['host']
        assert conn.port == mysql_params['port']
    assert conn.user == mysql_params['user']
    assert conn.db == mysql_params['db']
    assert conn.echo is False
    conn.close()


@pytest.mark.run_loop
async def test_connection_double_ensure_closed(connection_creator):
    conn = await connection_creator()
    assert conn.closed is False
    await conn.ensure_closed()
    assert conn.closed is True
    await conn.ensure_closed()
    assert conn.closed is True


@pytest.mark.run_loop
@pytest.mark.usefixtures("disable_gc")
async def test___del__(connection_creator):
    conn = await connection_creator()
    with pytest.warns(ResourceWarning):
        del conn
        gc.collect()


@pytest.mark.run_loop
async def test_previous_cursor_not_closed(connection_creator):
    conn = await connection_creator()
    cur1 = await conn.cursor()
    await cur1.execute("SELECT 1; SELECT 2")
    cur2 = await conn.cursor()
    await cur2.execute("SELECT 3;")
    resp = await cur2.fetchone()
    assert resp[0] == 3


@pytest.mark.run_loop
async def test_commit_during_multi_result(connection_creator):
    conn = await connection_creator()
    cur = await conn.cursor()
    await cur.execute("SELECT 1; SELECT 2;")
    await conn.commit()
    await cur.execute("SELECT 3;")
    resp = await cur.fetchone()
    assert resp[0] == 3



import builtins
import os
from unittest.mock import patch, MagicMock

import pytest
from pymysql.err import OperationalError


@pytest.fixture
def table_local_file(connection, loop):

    async def prepare_table(conn):
        c = await conn.cursor()
        await c.execute("DROP TABLE IF EXISTS test_load_local;")
        await c.execute("CREATE TABLE test_load_local "
                        "(a INTEGER, b INTEGER)")
        await c.close()

    async def drop_table(conn):
        c = await conn.cursor()
        await c.execute("DROP TABLE test_load_local")
        await c.close()

    loop.run_until_complete(prepare_table(connection))
    yield
    loop.run_until_complete(drop_table(connection))


@pytest.mark.run_loop
async def test_no_file(cursor, table_local_file):
    # Test load local infile when the file does not exist
    sql = "LOAD DATA LOCAL INFILE 'no_data.txt'" + \
          " INTO TABLE test_load_local fields " + \
          "terminated by ','"
    with pytest.raises(OperationalError):
        await cursor.execute(sql)


@pytest.mark.run_loop
async def test_error_on_file_read(cursor, table_local_file):

    with patch.object(builtins, 'open') as open_mocked:
        m = MagicMock()
        m.read.side_effect = OperationalError(1024, 'Error reading file')
        m.close.return_value = None
        open_mocked.return_value = m

        with pytest.raises(OperationalError):
            await cursor.execute("LOAD DATA LOCAL INFILE 'some.txt'"
                                 " INTO TABLE test_load_local fields "
                                 "terminated by ','")


@pytest.mark.run_loop
async def test_load_file(cursor, table_local_file):
    # Test load local infile with a valid file
    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'fixtures',
                            'load_local_data.txt')
    await cursor.execute(
        ("LOAD DATA LOCAL INFILE '{0}' INTO TABLE " +
         "test_load_local FIELDS TERMINATED BY ','").format(filename)
    )
    await cursor.execute("SELECT COUNT(*) FROM test_load_local")
    resp = await cursor.fetchone()
    assert 22749 == resp[0]


@pytest.mark.run_loop
async def test_load_warnings(cursor, table_local_file):
    # Test load local infile produces the appropriate warnings
    import warnings

    # TODO: Move to pathlib
    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'fixtures', 'load_local_warn_data.txt')

    sql = ("LOAD DATA LOCAL INFILE '{0}' INTO TABLE " +
           "test_load_local FIELDS TERMINATED BY ','").format(filename)

    with warnings.catch_warnings(record=True) as w:
        await cursor.execute(sql)
    assert "Incorrect integer value" in str(w[-1].message)



import datetime

import pytest
from sqlalchemy import MetaData, Table, Column, Integer, String
from sqlalchemy import func, DateTime, Boolean

from aiomysql import sa

meta = MetaData()
table = Table('sa_tbl_default_test', meta,
              Column('id', Integer, nullable=False, primary_key=True),
              Column('string_length', Integer,
                     default=func.length('qwerty')),
              Column('number', Integer, default=100, nullable=False),
              Column('description', String(255), nullable=False,
                     default='default test'),
              Column('created_at', DateTime,
                     default=datetime.datetime.now),
              Column('enabled', Boolean, default=True))


@pytest.fixture()
def make_engine(mysql_params, connection):
    async def _make_engine(**kwargs):
        if "unix_socket" in mysql_params:
            conn_args = {"unix_socket": mysql_params["unix_socket"]}
        else:
            conn_args = {
                "host": mysql_params['host'],
                "port": mysql_params['port'],
            }

        return (await sa.create_engine(db=mysql_params['db'],
                                       user=mysql_params['user'],
                                       password=mysql_params['password'],
                                       minsize=10,
                                       **conn_args,
                                       **kwargs))

    return _make_engine


async def start(engine):
    async with engine.acquire() as conn:
        await conn.execute("DROP TABLE IF EXISTS sa_tbl_default_test")
        await conn.execute("CREATE TABLE sa_tbl_default_test "
                           "(id integer,"
                           " string_length integer, "
                           "number integer,"
                           " description VARCHAR(255), "
                           "created_at DATETIME(6), "
                           "enabled TINYINT)")


@pytest.mark.run_loop
async def test_default_fields(make_engine):
    engine = await make_engine()
    await start(engine)
    async with engine.acquire() as conn:
        await conn.execute(table.insert().values())
        res = await conn.execute(table.select())
        row = await res.fetchone()
        assert row.string_length == 6
        assert row.number == 100
        assert row.description == 'default test'
        assert row.enabled is True
        assert type(row.created_at) == datetime.datetime


@pytest.mark.run_loop
async def test_default_fields_isnull(make_engine):
    engine = await make_engine()
    await start(engine)
    async with engine.acquire() as conn:
        created_at = None
        enabled = False
        await conn.execute(table.insert().values(
            enabled=enabled,
            created_at=created_at,
        ))

        res = await conn.execute(table.select())
        row = await res.fetchone()
        assert row.number == 100
        assert row.string_length == 6
        assert row.description == 'default test'
        assert row.enabled == enabled
        assert row.created_at == created_at


@pytest.mark.run_loop
async def test_default_fields_edit(make_engine):
    engine = await make_engine()
    await start(engine)
    async with engine.acquire() as conn:
        created_at = datetime.datetime.now()
        description = 'new descr'
        enabled = False
        number = 111
        await conn.execute(table.insert().values(
            description=description,
            enabled=enabled,
            created_at=created_at,
            number=number,
        ))

        res = await conn.execute(table.select())
        row = await res.fetchone()
        assert row.number == number
        assert row.string_length == 6
        assert row.description == description
        assert row.enabled == enabled
        assert row.created_at == created_at



import asyncio
import aiomysql


async def test_example(loop):
    pool = await aiomysql.create_pool(host='127.0.0.1', port=3306,
                                      user='root', password='',
                                      db='mysql', loop=loop)
    async with pool.acquire() as conn:
        async with conn.cursor() as cur:
            await cur.execute("SELECT 42;")
            print(cur.description)
            (r,) = await cur.fetchone()
            assert r == 42
    pool.close()
    await pool.wait_closed()


loop = asyncio.get_event_loop()
loop.run_until_complete(test_example(loop))



import asyncio
import aiomysql


loop = asyncio.get_event_loop()


@asyncio.coroutine
def test_example():
    pool = yield from aiomysql.create_pool(
        host='127.0.0.1',
        port=3306,
        user='root',
        password='',
        db='mysql',
        loop=loop
    )
    with (yield from pool) as conn:
        cur = yield from conn.cursor()
        yield from cur.execute("SELECT 10")
        # print(cur.description)
        (r,) = yield from cur.fetchone()
        assert r == 10
    pool.close()
    yield from pool.wait_closed()


loop.run_until_complete(test_example())



import asyncio
import gc
import os
import re
import ssl
import sys

import aiomysql
import pymysql
import pytest


# version gate can be removed once uvloop supports python 3.11
# https://github.com/MagicStack/uvloop/issues/450
# https://github.com/MagicStack/uvloop/pull/459
PY_311 = sys.version_info >= (3, 11)
if PY_311:
    uvloop = None
else:
    import uvloop


@pytest.fixture
def disable_gc():
    gc_enabled = gc.isenabled()
    if gc_enabled:
        gc.disable()
        gc.collect()
    yield
    if gc_enabled:
        gc.collect()
        gc.enable()


def pytest_generate_tests(metafunc):
    if 'loop_type' in metafunc.fixturenames:
        loop_type = ['asyncio', 'uvloop'] if uvloop else ['asyncio']
        metafunc.parametrize("loop_type", loop_type)

    if "mysql_address" in metafunc.fixturenames:
        mysql_addresses = []
        ids = []

        opt_mysql_unix_socket = \
            list(metafunc.config.getoption("mysql_unix_socket"))
        for i in range(len(opt_mysql_unix_socket)):
            if "=" in opt_mysql_unix_socket[i]:
                label, path = opt_mysql_unix_socket[i].split("=", 1)
                mysql_addresses.append(path)
                ids.append(label)
            else:
                mysql_addresses.append(opt_mysql_unix_socket[i])
                ids.append("unix{}".format(i))

        opt_mysql_address = list(metafunc.config.getoption("mysql_address"))
        for i in range(len(opt_mysql_address)):
            if "=" in opt_mysql_address[i]:
                label, addr = opt_mysql_address[i].split("=", 1)
                ids.append(label)
            else:
                addr = opt_mysql_address[i]
                ids.append("tcp{}".format(i))

            if ":" in addr:
                addr = addr.split(":", 1)
                mysql_addresses.append((addr[0], int(addr[1])))
            else:
                mysql_addresses.append((addr, 3306))

        # default to connecting to localhost
        if len(mysql_addresses) == 0:
            mysql_addresses = [("127.0.0.1", 3306)]
            ids = ["tcp-local"]

        assert len(mysql_addresses) == len(set(mysql_addresses)), \
            "mysql targets are not unique"
        assert len(ids) == len(set(ids)), \
            "mysql target names are not unique"

        metafunc.parametrize("mysql_address",
                             mysql_addresses,
                             ids=ids,
                             scope="session",
                             )


@pytest.fixture
def loop(request, loop_type):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)

    if uvloop and loop_type == 'uvloop':
        loop = uvloop.new_event_loop()
    else:
        loop = asyncio.new_event_loop()

    yield loop

    if not loop._closed:
        loop.call_soon(loop.stop)
        loop.run_forever()
        loop.close()
    gc.collect()
    asyncio.set_event_loop(None)


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name):
        if not callable(obj):
            return
        item = pytest.Function.from_parent(collector, name=name)
        if 'run_loop' in item.keywords:
            return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    """
    Run asyncio marked test functions in an event loop instead of a normal
    function call.
    """
    if 'run_loop' in pyfuncitem.keywords:
        funcargs = pyfuncitem.funcargs
        loop = funcargs['loop']
        testargs = {arg: funcargs[arg]
                    for arg in pyfuncitem._fixtureinfo.argnames}
        loop.run_until_complete(pyfuncitem.obj(**testargs))
        return True


def pytest_runtest_setup(item):
    if 'run_loop' in item.keywords and 'loop' not in item.fixturenames:
        # inject an event loop fixture for all async tests
        item.fixturenames.append('loop')


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "run_loop"
    )
    config.addinivalue_line(
        "markers",
        "mysql_version(db, version): run only on specific database versions"
    )


def pytest_addoption(parser):
    parser.addoption(
        "--mysql-address",
        action="append",
        default=[],
        help="list of addresses to connect to: [name=]host[:port]",
    )
    parser.addoption(
        "--mysql-unix-socket",
        action="append",
        default=[],
        help="list of unix sockets to connect to: [name=]/path/to/socket",
    )


@pytest.fixture
def mysql_params(mysql_server):
    params = {**mysql_server['conn_params'],
              "db": os.environ.get('MYSQL_DB', 'test_pymysql'),
              "local_infile": True,
              "use_unicode": True,
              }
    return params


# TODO: fix this workaround
async def _cursor_wrapper(conn):
    return await conn.cursor()


@pytest.fixture
def cursor(connection, loop):
    cur = loop.run_until_complete(_cursor_wrapper(connection))
    yield cur
    loop.run_until_complete(cur.close())


@pytest.fixture
def connection(mysql_params, loop):
    coro = aiomysql.connect(loop=loop, **mysql_params)
    conn = loop.run_until_complete(coro)
    yield conn
    loop.run_until_complete(conn.ensure_closed())


@pytest.fixture
def connection_creator(mysql_params, loop):
    connections = []

    async def f(**kw):
        conn_kw = mysql_params.copy()
        conn_kw.update(kw)
        _loop = conn_kw.pop('loop', loop)
        conn = await aiomysql.connect(loop=_loop, **conn_kw)
        connections.append(conn)
        return conn

    yield f

    for conn in connections:
        try:
            loop.run_until_complete(conn.ensure_closed())
        except ConnectionResetError:
            pass


@pytest.fixture
def pool_creator(mysql_params, loop):
    pools = []

    async def f(**kw):
        conn_kw = mysql_params.copy()
        conn_kw.update(kw)
        _loop = conn_kw.pop('loop', loop)
        pool = await aiomysql.create_pool(loop=_loop, **conn_kw)
        pools.append(pool)
        return pool

    yield f

    for pool in pools:
        pool.close()
        loop.run_until_complete(pool.wait_closed())


@pytest.fixture
def table_cleanup(loop, connection):
    table_list = []
    cursor = loop.run_until_complete(_cursor_wrapper(connection))

    def _register_table(table_name):
        table_list.append(table_name)

    yield _register_table
    for t in table_list:
        # TODO: probably this is not safe code
        sql = "DROP TABLE IF EXISTS {};".format(t)
        loop.run_until_complete(cursor.execute(sql))


@pytest.fixture(scope='session')
def mysql_server(mysql_address):
    unix_socket = type(mysql_address) is str

    if not unix_socket:
        ssl_directory = os.path.join(os.path.dirname(__file__),
                                     'ssl_resources', 'ssl')
        ca_file = os.path.join(ssl_directory, 'ca.pem')

        ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        ctx.check_hostname = False
        ctx.load_verify_locations(cafile=ca_file)
        # ctx.verify_mode = ssl.CERT_NONE

    server_params = {
        'user': 'root',
        'password': os.environ.get("MYSQL_ROOT_PASSWORD"),
    }

    if unix_socket:
        server_params["unix_socket"] = mysql_address
    else:
        server_params["host"] = mysql_address[0]
        server_params["port"] = mysql_address[1]
        server_params["ssl"] = ctx

    try:
        connection = pymysql.connect(
            db='mysql',
            charset='utf8mb4',
            cursorclass=pymysql.cursors.DictCursor,
            **server_params)

        with connection.cursor() as cursor:
            cursor.execute("SELECT VERSION() AS version")
            server_version = cursor.fetchone()["version"]
            server_version_tuple = tuple(
                (int(dig) if dig is not None else 0)
                for dig in
                re.match(r"^(\d+)\.(\d+)(?:\.(\d+))?", server_version).group(1, 2, 3)
            )
            server_version_tuple_short = (server_version_tuple[0],
                                          server_version_tuple[1])
            if server_version_tuple_short in [(5, 7), (8, 0)]:
                db_type = "mysql"
            elif server_version_tuple[0] == 10:
                db_type = "mariadb"
            else:
                pytest.fail("Unable to determine database type from {!r}"
                            .format(server_version_tuple))

            if not unix_socket:
                cursor.execute("SHOW VARIABLES LIKE '%ssl%';")

                result = cursor.fetchall()
                result = {item['Variable_name']:
                          item['Value'] for item in result}

                assert result['have_ssl'] == "YES", \
                    "SSL Not Enabled on MySQL"

                cursor.execute("SHOW STATUS LIKE 'Ssl_version%'")

                result = cursor.fetchone()
                # As we connected with TLS, it should start with that :D
                assert result['Value'].startswith('TLS'), \
                    "Not connected to the database with TLS"

            # Drop possibly existing old databases
            cursor.execute('DROP DATABASE IF EXISTS test_pymysql;')
            cursor.execute('DROP DATABASE IF EXISTS test_pymysql2;')

            # Create Databases
            cursor.execute('CREATE DATABASE test_pymysql  '
                           'DEFAULT CHARACTER SET utf8 '
                           'DEFAULT COLLATE utf8_general_ci;')
            cursor.execute('CREATE DATABASE test_pymysql2 '
                           'DEFAULT CHARACTER SET utf8 '
                           'DEFAULT COLLATE utf8_general_ci;')

            # Do MySQL8+ Specific Setup
            if db_type == "mysql" and server_version_tuple_short == (8, 0):
                # Drop existing users
                cursor.execute('DROP USER IF EXISTS user_sha256;')
                cursor.execute('DROP USER IF EXISTS nopass_sha256;')
                cursor.execute('DROP USER IF EXISTS user_caching_sha2;')
                cursor.execute('DROP USER IF EXISTS nopass_caching_sha2;')

                # Create Users to test SHA256
                cursor.execute('CREATE USER user_sha256 '
                               'IDENTIFIED WITH "sha256_password" '
                               'BY "pass_sha256"')
                cursor.execute('CREATE USER nopass_sha256 '
                               'IDENTIFIED WITH "sha256_password"')
                cursor.execute('CREATE USER user_caching_sha2   '
                               'IDENTIFIED '
                               'WITH "caching_sha2_password" '
                               'BY "pass_caching_sha2"')
                cursor.execute('CREATE USER nopass_caching_sha2 '
                               'IDENTIFIED '
                               'WITH "caching_sha2_password" '
                               'PASSWORD EXPIRE NEVER')
                cursor.execute('FLUSH PRIVILEGES')
    except Exception:
        pytest.fail("Cannot initialize MySQL environment")

    return {
        "conn_params": server_params,
        "server_version": server_version,
        "server_version_tuple": server_version_tuple,
        "server_version_tuple_short": server_version_tuple_short,
        "db_type": db_type,
    }



import pytest

from aiomysql import SSCursor
from aiomysql import sa

from sqlalchemy import MetaData, Table, Column, Integer, String


meta = MetaData()
tbl = Table('tbl', meta,
            Column('id', Integer, nullable=False, primary_key=True),
            Column('name', String(255)))


@pytest.fixture
def table(loop, connection_creator, table_cleanup):
    async def f():
        connection = await connection_creator()
        cursor = await connection.cursor()
        await cursor.execute("DROP TABLE IF EXISTS tbl;")
        await cursor.execute("""CREATE TABLE tbl (
                 id MEDIUMINT NOT NULL AUTO_INCREMENT,
                 name VARCHAR(255) NOT NULL,
                 PRIMARY KEY (id));""")

        for i in [(1, 'a'), (2, 'b'), (3, 'c')]:
            await cursor.execute("INSERT INTO tbl VALUES(%s, %s)", i)

        await cursor.execute("commit;")
        await cursor.close()

    table_cleanup('tbl')
    loop.run_until_complete(f())


@pytest.mark.run_loop
async def test_async_cursor(cursor, table):
    ret = []
    await cursor.execute('SELECT * from tbl;')
    async for i in cursor:
        ret.append(i)
    assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret


@pytest.mark.run_loop
async def test_async_cursor_server_side(connection, table):
    ret = []
    cursor = await connection.cursor(SSCursor)
    await cursor.execute('SELECT * from tbl;')
    async for i in cursor:
        ret.append(i)
    assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret


@pytest.mark.run_loop
async def test_async_iter_over_sa_result(mysql_params, table, loop):
    ret = []
    engine = await sa.create_engine(**mysql_params, loop=loop)
    conn = await engine.acquire()

    async for i in (await conn.execute(tbl.select())):
        ret.append(i)

    assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret
    engine.terminate()



import asyncio
import sqlalchemy as sa

from aiomysql.sa import create_engine


metadata = sa.MetaData()

tbl = sa.Table('tbl', metadata,
               sa.Column('id', sa.Integer, primary_key=True),
               sa.Column('val', sa.String(255)))


async def create_table(engine):
    async with engine.acquire() as conn:
        await conn.execute('DROP TABLE IF EXISTS tbl')
        await conn.execute('''CREATE TABLE tbl (
                              id serial PRIMARY KEY,
                              val varchar(255))''')


async def go(loop):
    engine = await create_engine(user='root', db='test_pymysql',
                                 host='127.0.0.1', password='', loop=loop)
    await create_table(engine)
    async with engine.acquire() as conn:
        await conn.execute(tbl.insert().values(val='abc'))
        await conn.execute(tbl.insert().values(val='xyz'))

        async for row in conn.execute(tbl.select()):
            print(row.id, row.val)

        await conn.execute("commit")

    engine.close()
    await engine.wait_closed()


loop = asyncio.get_event_loop()
loop.run_until_complete(go(loop))



import asyncio
import unittest


class BaseTest(unittest.TestCase):
    """Base test case for unittests.
    """
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

    def tearDown(self):
        self.loop.close()
        del self.loop



import asyncio

import pytest
from pymysql import NotSupportedError

from aiomysql import ProgrammingError, InterfaceError, OperationalError
from aiomysql.cursors import SSCursor


DATA = [
    ('America', '', 'America/Jamaica'),
    ('America', '', 'America/Los_Angeles'),
    ('America', '', 'America/Lima'),
    ('America', '', 'America/New_York'),
    ('America', '', 'America/Menominee'),
    ('America', '', 'America/Havana'),
    ('America', '', 'America/El_Salvador'),
    ('America', '', 'America/Costa_Rica'),
    ('America', '', 'America/Denver'),
    ('America', '', 'America/Detroit'), ]


async def _prepare(conn):
    cursor = await conn.cursor()
    # Create table
    await cursor.execute('DROP TABLE IF EXISTS tz_data;')
    await cursor.execute('CREATE TABLE tz_data ('
                         'region VARCHAR(64),'
                         'zone VARCHAR(64),'
                         'name VARCHAR(64))')

    await cursor.executemany(
        'INSERT INTO tz_data VALUES (%s, %s, %s)', DATA)
    await conn.commit()
    await cursor.close()


@pytest.mark.run_loop
async def test_ssursor(connection):
    # affected_rows = 18446744073709551615
    conn = connection
    cursor = await conn.cursor(SSCursor)
    # Create table
    await cursor.execute('DROP TABLE IF EXISTS tz_data;')
    await cursor.execute(('CREATE TABLE tz_data ('
                          'region VARCHAR(64),'
                          'zone VARCHAR(64),'
                          'name VARCHAR(64))'))

    # Test INSERT
    for i in DATA:
        await cursor.execute(
            'INSERT INTO tz_data VALUES (%s, %s, %s)', i)
        assert conn.affected_rows() == 1, 'affected_rows does not match'
    await conn.commit()

    # Test update, affected_rows()
    await cursor.execute('UPDATE tz_data SET zone = %s', ['Foo'])
    await conn.commit()

    assert cursor.rowcount == len(DATA), \
        'Update failed. affected_rows != %s' % (str(len(DATA)))

    await cursor.close()
    await cursor.close()


@pytest.mark.run_loop
async def test_sscursor_fetchall(connection):
    conn = connection
    cursor = await conn.cursor(SSCursor)

    await _prepare(conn)
    await cursor.execute('SELECT * FROM tz_data')
    fetched_data = await cursor.fetchall()
    assert len(fetched_data) == len(DATA), \
        'fetchall failed. Number of rows does not match'


@pytest.mark.run_loop
async def test_sscursor_fetchmany(connection):
    conn = connection
    cursor = await conn.cursor(SSCursor)
    await _prepare(conn)
    await cursor.execute('SELECT * FROM tz_data')
    fetched_data = await cursor.fetchmany(2)
    assert len(fetched_data) == 2, \
        'fetchmany failed. Number of rows does not match'

    await cursor.close()
    # test default fetchmany size
    cursor = await conn.cursor(SSCursor)
    await cursor.execute('SELECT * FROM tz_data;')
    fetched_data = await cursor.fetchmany()
    assert len(fetched_data) == 1


@pytest.mark.run_loop
async def test_sscursor_executemany(connection):
    conn = connection
    await _prepare(conn)
    cursor = await conn.cursor(SSCursor)
    # Test executemany
    await cursor.executemany(
        'INSERT INTO tz_data VALUES (%s, %s, %s)', DATA)
    msg = 'executemany failed. cursor.rowcount != %s'
    assert cursor.rowcount == len(DATA), msg % (str(len(DATA)))


@pytest.mark.run_loop
async def test_sscursor_scroll_relative(connection):
    conn = connection
    await _prepare(conn)
    cursor = await conn.cursor(SSCursor)
    await cursor.execute('SELECT * FROM tz_data;')
    await cursor.scroll(1)
    ret = await cursor.fetchone()
    assert ('America', '', 'America/Los_Angeles') == ret


@pytest.mark.run_loop
async def test_sscursor_scroll_absolute(connection):
    conn = connection
    await _prepare(conn)
    cursor = await conn.cursor(SSCursor)
    await cursor.execute('SELECT * FROM tz_data;')
    await cursor.scroll(2, mode='absolute')
    ret = await cursor.fetchone()
    assert ('America', '', 'America/Lima') == ret


@pytest.mark.run_loop
async def test_sscursor_scroll_errors(connection):
    conn = connection
    await _prepare(conn)
    cursor = await conn.cursor(SSCursor)

    await cursor.execute('SELECT * FROM tz_data;')

    with pytest.raises(NotSupportedError):
        await cursor.scroll(-2, mode='relative')

    await cursor.scroll(2, mode='absolute')

    with pytest.raises(NotSupportedError):
        await cursor.scroll(1, mode='absolute')
    with pytest.raises(ProgrammingError):
        await cursor.scroll(2, mode='not_valid_mode')


@pytest.mark.run_loop
async def test_sscursor_cancel(connection):
    conn = connection
    cur = await conn.cursor(SSCursor)
    # Prepare A LOT of data

    await cur.execute('DROP TABLE IF EXISTS long_seq;')
    await cur.execute(
        """ CREATE TABLE long_seq (
              id int(11)
            )
        """)

    ids = [(x) for x in range(100000)]
    await cur.executemany('INSERT INTO long_seq VALUES (%s)', ids)

    # Will return several results. All we need at this point
    big_str = "x" * 10000
    await cur.execute(
        """SELECT '{}' as id FROM long_seq;
        """.format(big_str))
    first = await cur.fetchone()
    assert first == (big_str,)

    async def read_cursor():
        while True:
            res = await cur.fetchone()
            if res is None:
                break
    task = asyncio.ensure_future(read_cursor())
    await asyncio.sleep(0)
    assert not task.done(), "Test failed to produce needed condition."
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass

    with pytest.raises(InterfaceError):
        await conn.cursor(SSCursor)


@pytest.mark.run_loop
async def test_sscursor_discarded_result(connection):
    conn = connection
    await _prepare(conn)
    async with conn.cursor(SSCursor) as cursor:
        await cursor.execute("select 1")
        await cursor.execute("select 2")
        ret = await cursor.fetchone()
    assert (2,) == ret


@pytest.mark.run_loop
async def test_max_execution_time(mysql_server, connection):
    conn = connection

    async with connection.cursor() as cur:
        await cur.execute("DROP TABLE IF EXISTS tbl;")

        await cur.execute(
            """
            CREATE TABLE tbl (
            id MEDIUMINT NOT NULL AUTO_INCREMENT,
            name VARCHAR(255) NOT NULL,
            PRIMARY KEY (id));
            """
        )

        for i in [(1, "a"), (2, "b"), (3, "c")]:
            await cur.execute("INSERT INTO tbl VALUES(%s, %s)", i)

        await conn.commit()

    async with conn.cursor(SSCursor) as cur:
        # MySQL MAX_EXECUTION_TIME takes ms
        # MariaDB max_statement_time takes seconds as int/float, introduced in 10.1

        # this will sleep 0.01 seconds per row
        if mysql_server["db_type"] == "mysql":
            sql = """
                  SELECT /*+ MAX_EXECUTION_TIME(2000) */
                  name, sleep(0.01) FROM tbl
                  """
        else:
            sql = """
                  SET STATEMENT max_statement_time=2 FOR
                  SELECT name, sleep(0.01) FROM tbl
                  """

        await cur.execute(sql)
        # unlike Cursor, SSCursor returns a list of tuples here

        assert (await cur.fetchall()) == [
            ("a", 0),
            ("b", 0),
            ("c", 0),
        ]

        if mysql_server["db_type"] == "mysql":
            sql = """
                      SELECT /*+ MAX_EXECUTION_TIME(2000) */
                      name, sleep(0.01) FROM tbl
                      """
        else:
            sql = """
                      SET STATEMENT max_statement_time=2 FOR
                      SELECT name, sleep(0.01) FROM tbl
                      """
        await cur.execute(sql)
        assert (await cur.fetchone()) == ("a", 0)

        # this discards the previous unfinished query and raises an
        # incomplete unbuffered query warning
        with pytest.warns(UserWarning):
            await cur.execute("SELECT 1")
        assert (await cur.fetchone()) == (1,)

        # SSCursor will not read the EOF packet until we try to read
        # another row. Skipping this will raise an incomplete unbuffered
        # query warning in the next cur.execute().
        assert (await cur.fetchone()) is None

        if mysql_server["db_type"] == "mysql":
            sql = """
                      SELECT /*+ MAX_EXECUTION_TIME(1) */
                      name, sleep(1) FROM tbl
                      """
        else:
            sql = """
                      SET STATEMENT max_statement_time=0.001 FOR
                      SELECT name, sleep(1) FROM tbl
                      """
        with pytest.raises(OperationalError) as cm:
            # in an unbuffered cursor the OperationalError may not show up
            # until fetching the entire result
            await cur.execute(sql)
            await cur.fetchall()

        if mysql_server["db_type"] == "mysql":
            # this constant was only introduced in MySQL 5.7, not sure
            # what was returned before, may have been ER_QUERY_INTERRUPTED

            # this constant is pending a new PyMySQL release
            # assert cm.value.args[0] == pymysql.constants.ER.QUERY_TIMEOUT
            assert cm.value.args[0] == 3024
        else:
            # this constant is pending a new PyMySQL release
            # assert cm.value.args[0] == pymysql.constants.ER.STATEMENT_TIMEOUT
            assert cm.value.args[0] == 1969

        # connection should still be fine at this point
        await cur.execute("SELECT 1")
        assert (await cur.fetchone()) == (1,)



import asyncio
import ssl
import aiomysql

ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
ctx.check_hostname = False
ctx.load_verify_locations(cafile='../tests/ssl_resources/ssl/ca.pem')


async def main():
    async with aiomysql.create_pool(
            host='localhost', port=3306, user='root',
            password='rootpw', ssl=ctx,
            auth_plugin='mysql_clear_password') as pool:

        async with pool.get() as conn:
            async with conn.cursor() as cur:
                # Run simple command
                await cur.execute("SHOW DATABASES;")
                value = await cur.fetchall()

                values = [item[0] for item in value]
                # Spot check the answers, we should at least have mysql
                # and information_schema
                assert 'mysql' in values, \
                    'Could not find the "mysql" table'
                assert 'information_schema' in values, \
                    'Could not find the "mysql" table'

                # Check TLS variables
                await cur.execute("SHOW STATUS LIKE 'Ssl_version%';")
                value = await cur.fetchone()

                # The context has TLS
                assert value[1].startswith('TLS'), \
                    'Not connected to the database with TLS'

asyncio.get_event_loop().run_until_complete(main())



import copy

import aiomysql.cursors

import pytest


BOB = ("bob", 21, {"k1": "pretty", "k2": [18, 25]})
JIM = ("jim", 56, {"k1": "rich", "k2": [20, 60]})
FRED = ("fred", 100, {"k1": "longevity", "k2": [100, 160]})


@pytest.fixture()
async def prepare(connection):

    havejson = True

    c = await connection.cursor(aiomysql.cursors.DeserializationCursor)

    # create a table ane some data to query
    await c.execute("drop table if exists deserialize_cursor")
    await c.execute("select VERSION()")
    v = await c.fetchone()
    version, *db_type = v[0].split('-', 1)
    version = float(".".join(version.split('.', 2)[:2]))
    ismariadb = db_type and 'mariadb' in db_type[0].lower()
    if ismariadb or version < 5.7:
        await c.execute(
            """CREATE TABLE deserialize_cursor
             (name char(20), age int , claim text)""")
        havejson = False
    else:
        await c.execute(
            """CREATE TABLE deserialize_cursor
             (name char(20), age int , claim json)""")
    data = [("bob", 21, '{"k1": "pretty", "k2": [18, 25]}'),
            ("jim", 56, '{"k1": "rich", "k2": [20, 60]}'),
            ("fred", 100, '{"k1": "longevity", "k2": [100, 160]}')]
    await c.executemany("insert into deserialize_cursor values "
                        "(%s,%s,%s)", data)

    return havejson


@pytest.mark.run_loop
async def test_deserialize_cursor(prepare, connection):
    havejson = await prepare
    if not havejson:
        return
    bob, jim, fred = copy.deepcopy(BOB), copy.deepcopy(
        JIM), copy.deepcopy(FRED)
    # all assert test compare to the structure as would come
    # out from MySQLdb
    conn = connection
    c = await conn.cursor(aiomysql.cursors.DeserializationCursor)

    # pull back the single row dict for bob and check
    await c.execute("SELECT * from deserialize_cursor "
                    "where name='bob'")
    r = await c.fetchone()
    assert bob == r, "fetchone via DeserializeCursor failed"
    # same again, but via fetchall => tuple)
    await c.execute("SELECT * from deserialize_cursor "
                    "where name='bob'")
    r = await c.fetchall()
    assert [bob] == r, \
        "fetch a 1 row result via fetchall failed via DeserializeCursor"
    # get all 3 row via fetchall
    await c.execute("SELECT * from deserialize_cursor")
    r = await c.fetchall()
    assert [bob, jim, fred] == r, "fetchall failed via DictCursor"

    # get all 2 row via fetchmany
    await c.execute("SELECT * from deserialize_cursor")
    r = await c.fetchmany(2)
    assert [bob, jim] == r, "fetchmany failed via DictCursor"
    await c.execute('commit')


@pytest.mark.run_loop
async def test_deserialize_cursor_low_version(prepare, connection):
    havejson = await prepare
    if havejson:
        return
    bob = ("bob", 21, '{"k1": "pretty", "k2": [18, 25]}')
    jim = ("jim", 56, '{"k1": "rich", "k2": [20, 60]}')
    fred = ("fred", 100, '{"k1": "longevity", "k2": [100, 160]}')
    # all assert test compare to the structure as would come
    # out from MySQLdb
    conn = connection
    c = await conn.cursor(aiomysql.cursors.DeserializationCursor)

    # pull back the single row dict for bob and check
    await c.execute("SELECT * from deserialize_cursor where name='bob'")
    r = await c.fetchone()
    assert bob == r, "fetchone via DeserializeCursor failed"
    # same again, but via fetchall => tuple)
    await c.execute("SELECT * from deserialize_cursor "
                    "where name='bob'")
    r = await c.fetchall()
    assert [bob] == r, \
        "fetch a 1 row result via fetchall failed via DeserializeCursor"
    # get all 3 row via fetchall
    await c.execute("SELECT * from deserialize_cursor")
    r = await c.fetchall()
    assert [bob, jim, fred] == r, "fetchall failed via DictCursor"

    # get all 2 row via fetchmany
    await c.execute("SELECT * from deserialize_cursor")
    r = await c.fetchmany(2)
    assert [bob, jim] == r, "fetchmany failed via DictCursor"
    await c.execute('commit')


@pytest.mark.run_loop
async def test_deserializedictcursor(prepare, connection):
    havejson = await prepare
    if not havejson:
        return
    bob = {'name': 'bob', 'age': 21,
           'claim': {"k1": "pretty", "k2": [18, 25]}}
    conn = connection
    c = await conn.cursor(aiomysql.cursors.DeserializationCursor,
                          aiomysql.cursors.DictCursor)
    await c.execute("SELECT * from deserialize_cursor "
                    "where name='bob'")
    r = await c.fetchall()
    assert [bob] == r, \
        "fetch a 1 row result via fetchall failed via DeserializationCursor"


@pytest.mark.run_loop
async def test_ssdeserializecursor(prepare, connection):
    havejson = await prepare
    if not havejson:
        return
    conn = connection
    c = await conn.cursor(aiomysql.cursors.SSCursor,
                          aiomysql.cursors.DeserializationCursor)
    await c.execute("SELECT * from deserialize_cursor "
                    "where name='bob'")
    r = await c.fetchall()
    assert [BOB] == r, \
        "fetch a 1 row result via fetchall failed via DeserializationCursor"


@pytest.mark.run_loop
async def test_ssdeserializedictcursor(prepare, connection):
    havejson = await prepare
    if not havejson:
        return
    bob = {'name': 'bob', 'age': 21,
           'claim': {"k1": "pretty", "k2": [18, 25]}}
    conn = connection
    c = await conn.cursor(aiomysql.cursors.SSCursor,
                          aiomysql.cursors.DeserializationCursor,
                          aiomysql.cursors.DictCursor)
    await c.execute("SELECT * from deserialize_cursor "
                    "where name='bob'")
    r = await c.fetchall()
    assert [bob] == r, \
        "fetch a 1 row result via fetchall failed via DeserializationCursor"



import asyncio
import aiomysql


loop = asyncio.get_event_loop()


@asyncio.coroutine
def test_example_executemany():
    conn = yield from aiomysql.connect(host='127.0.0.1', port=3306,
                                       user='root', password='',
                                       db='test_pymysql', loop=loop)

    cur = yield from conn.cursor()
    yield from cur.execute("DROP TABLE IF EXISTS music_style;")
    yield from cur.execute("""CREATE TABLE music_style
                              (id INT,
                              name VARCHAR(255),
                              PRIMARY KEY (id));""")
    yield from conn.commit()

    # insert 3 rows one by one
    yield from cur.execute("INSERT INTO music_style VALUES(1,'heavy metal')")
    yield from cur.execute("INSERT INTO music_style VALUES(2,'death metal');")
    yield from cur.execute("INSERT INTO music_style VALUES(3,'power metal');")
    yield from conn.commit()

    # insert 3 row by one long query using *executemane* method
    data = [(4, 'gothic metal'), (5, 'doom metal'), (6, 'post metal')]
    yield from cur.executemany(
        "INSERT INTO music_style (id, name)"
        "values (%s,%s)", data)
    yield from conn.commit()

    # fetch all insert row from table music_style
    yield from cur.execute("SELECT * FROM music_style;")
    result = yield from cur.fetchall()
    print(result)

    yield from cur.close()
    conn.close()


loop.run_until_complete(test_example_executemany())



import asyncio
import aiomysql


loop = asyncio.get_event_loop()


@asyncio.coroutine
def test_example():
    conn = yield from aiomysql.connect(host='127.0.0.1', port=3306,
                                       user='root', password='',
                                       db='test_pymysql', loop=loop)

    cur = yield from conn.cursor()
    yield from cur.execute("DROP PROCEDURE IF EXISTS myinc;")
    yield from cur.execute("""CREATE PROCEDURE myinc(p1 INT)
                           BEGIN
                               SELECT p1 + 1;
                           END
                           """)

    yield from cur.callproc('myinc', [1])
    (ret, ) = yield from cur.fetchone()
    assert 2, ret
    print(ret)

    yield from cur.close()
    conn.close()


loop.run_until_complete(test_example())



import asyncio
import sqlalchemy as sa

from aiomysql.sa import create_engine


metadata = sa.MetaData()

tbl = sa.Table('tbl', metadata,
               sa.Column('id', sa.Integer, primary_key=True),
               sa.Column('val', sa.String(255)))


@asyncio.coroutine
def create_table(engine):
    with (yield from engine) as conn:
        yield from conn.execute('DROP TABLE IF EXISTS tbl')
        yield from conn.execute('''CREATE TABLE tbl (
                                            id serial PRIMARY KEY,
                                            val varchar(255))''')


@asyncio.coroutine
def go():
    engine = yield from create_engine(user='root',
                                      db='test_pymysql',
                                      host='127.0.0.1',
                                      password='')

    yield from create_table(engine)
    with (yield from engine) as conn:
        yield from conn.execute(tbl.insert().values(val='abc'))

        res = yield from conn.execute(tbl.select())
        for row in res:
            print(row.id, row.val)

        yield from conn.commit()

    engine.close()
    yield from engine.wait_closed()


asyncio.get_event_loop().run_until_complete(go())



import re
import json
import warnings
import contextlib

from pymysql.err import (
    Warning, Error, InterfaceError, DataError,
    DatabaseError, OperationalError, IntegrityError, InternalError,
    NotSupportedError, ProgrammingError)

from .log import logger
from .connection import FIELD_TYPE

# https://github.com/PyMySQL/PyMySQL/blob/master/pymysql/cursors.py#L11-L18

#: Regular expression for :meth:`Cursor.executemany`.
#: executemany only supports simple bulk insert.
#: You can use it to load large dataset.
RE_INSERT_VALUES = re.compile(
    r"\s*((?:INSERT|REPLACE)\s.+\sVALUES?\s+)" +
    r"(\(\s*(?:%s|%\(.+\)s)\s*(?:,\s*(?:%s|%\(.+\)s)\s*)*\))" +
    r"(\s*(?:ON DUPLICATE.*)?);?\s*\Z",
    re.IGNORECASE | re.DOTALL)


class Cursor:
    """Cursor is used to interact with the database."""

    #: Max statement size which :meth:`executemany` generates.
    #:
    #: Max size of allowed statement is max_allowed_packet -
    # packet_header_size.
    #: Default value of max_allowed_packet is 1048576.
    max_stmt_length = 1024000

    def __init__(self, connection, echo=False):
        """Do not create an instance of a Cursor yourself. Call
        connections.Connection.cursor().
        """
        self._connection = connection
        self._loop = self._connection.loop
        self._description = None
        self._rownumber = 0
        self._rowcount = -1
        self._arraysize = 1
        self._executed = None
        self._result = None
        self._rows = None
        self._lastrowid = None
        self._echo = echo

    @property
    def connection(self):
        """This read-only attribute return a reference to the Connection
        object on which the cursor was created."""
        return self._connection

    @property
    def description(self):
        """This read-only attribute is a sequence of 7-item sequences.

        Each of these sequences is a collections.namedtuple containing
        information describing one result column:

        0.  name: the name of the column returned.
        1.  type_code: the type of the column.
        2.  display_size: the actual length of the column in bytes.
        3.  internal_size: the size in bytes of the column associated to
            this column on the server.
        4.  precision: total number of significant digits in columns of
            type NUMERIC. None for other types.
        5.  scale: count of decimal digits in the fractional part in
            columns of type NUMERIC. None for other types.
        6.  null_ok: always None as not easy to retrieve from the libpq.

        This attribute will be None for operations that do not
        return rows or if the cursor has not had an operation invoked
        via the execute() method yet.
        """
        return self._description

    @property
    def rowcount(self):
        """Returns the number of rows that has been produced of affected.

        This read-only attribute specifies the number of rows that the
        last :meth:`execute` produced (for Data Query Language
        statements like SELECT) or affected (for Data Manipulation
        Language statements like UPDATE or INSERT).

        The attribute is -1 in case no .execute() has been performed
        on the cursor or the row count of the last operation if it
        can't be determined by the interface.
        """
        return self._rowcount

    @property
    def rownumber(self):
        """Row index.

        This read-only attribute provides the current 0-based index of the
        cursor in the result set or ``None`` if the index cannot be
        determined.
        """

        return self._rownumber

    @property
    def arraysize(self):
        """How many rows will be returned by fetchmany() call.

        This read/write attribute specifies the number of rows to
        fetch at a time with fetchmany(). It defaults to
        1 meaning to fetch a single row at a time.

        """
        return self._arraysize

    @arraysize.setter
    def arraysize(self, val):
        """How many rows will be returned by fetchmany() call.

        This read/write attribute specifies the number of rows to
        fetch at a time with fetchmany(). It defaults to
        1 meaning to fetch a single row at a time.

        """
        self._arraysize = val

    @property
    def lastrowid(self):
        """This read-only property returns the value generated for an
        AUTO_INCREMENT column by the previous INSERT or UPDATE statement
        or None when there is no such value available. For example,
        if you perform an INSERT into a table that contains an AUTO_INCREMENT
        column, lastrowid returns the AUTO_INCREMENT value for the new row.
        """
        return self._lastrowid

    @property
    def echo(self):
        """Return echo mode status."""
        return self._echo

    @property
    def closed(self):
        """The readonly property that returns ``True`` if connections was
        detached from current cursor
        """
        return True if not self._connection else False

    async def close(self):
        """Closing a cursor just exhausts all remaining data."""
        conn = self._connection
        if conn is None:
            return
        try:
            while (await self.nextset()):
                pass
        finally:
            self._connection = None

    def _get_db(self):
        if not self._connection:
            raise ProgrammingError("Cursor closed")
        return self._connection

    def _check_executed(self):
        if not self._executed:
            raise ProgrammingError("execute() first")

    def _conv_row(self, row):
        return row

    def setinputsizes(self, *args):
        """Does nothing, required by DB API."""

    def setoutputsizes(self, *args):
        """Does nothing, required by DB API."""

    async def nextset(self):
        """Get the next query set"""
        conn = self._get_db()
        current_result = self._result
        if current_result is None or current_result is not conn._result:
            return
        if not current_result.has_next:
            return
        self._result = None
        self._clear_result()
        await conn.next_result()
        await self._do_get_result()
        return True

    def _escape_args(self, args, conn):
        if isinstance(args, (tuple, list)):
            return tuple(conn.escape(arg) for arg in args)
        elif isinstance(args, dict):
            return dict((key, conn.escape(val)) for (key, val) in args.items())
        else:
            # If it's not a dictionary let's try escaping it anyways.
            # Worst case it will throw a Value error
            return conn.escape(args)

    def mogrify(self, query, args=None):
        """ Returns the exact string that is sent to the database by calling
        the execute() method. This method follows the extension to the DB
        API 2.0 followed by Psycopg.

        :param query: ``str`` sql statement
        :param args: ``tuple`` or ``list`` of arguments for sql query
        """
        conn = self._get_db()
        if args is not None:
            query = query % self._escape_args(args, conn)
        return query

    async def execute(self, query, args=None):
        """Executes the given operation

        Executes the given operation substituting any markers with
        the given parameters.

        For example, getting all rows where id is 5:
          cursor.execute("SELECT * FROM t1 WHERE id = %s", (5,))

        :param query: ``str`` sql statement
        :param args: ``tuple`` or ``list`` of arguments for sql query
        :returns: ``int``, number of rows that has been produced of affected
        """
        conn = self._get_db()

        while (await self.nextset()):
            pass

        if args is not None:
            query = query % self._escape_args(args, conn)

        await self._query(query)
        self._executed = query
        if self._echo:
            logger.info(query)
            logger.info("%r", args)
        return self._rowcount

    async def executemany(self, query, args):
        """Execute the given operation multiple times

        The executemany() method will execute the operation iterating
        over the list of parameters in seq_params.

        Example: Inserting 3 new employees and their phone number

            data = [
                ('Jane','555-001'),
                ('Joe', '555-001'),
                ('John', '555-003')
                ]
            stmt = "INSERT INTO employees (name, phone) VALUES ('%s','%s')"
            await cursor.executemany(stmt, data)

        INSERT or REPLACE statements are optimized by batching the data,
        that is using the MySQL multiple rows syntax.

        :param query: `str`, sql statement
        :param args: ``tuple`` or ``list`` of arguments for sql query
        """
        if not args:
            return

        if self._echo:
            logger.info("CALL %s", query)
            logger.info("%r", args)

        m = RE_INSERT_VALUES.match(query)
        if m:
            q_prefix = m.group(1) % ()
            q_values = m.group(2).rstrip()
            q_postfix = m.group(3) or ''
            assert q_values[0] == '(' and q_values[-1] == ')'
            return (await self._do_execute_many(
                q_prefix, q_values, q_postfix, args, self.max_stmt_length,
                self._get_db().encoding))
        else:
            rows = 0
            for arg in args:
                await self.execute(query, arg)
                rows += self._rowcount
            self._rowcount = rows
        return self._rowcount

    async def _do_execute_many(self, prefix, values, postfix, args,
                               max_stmt_length, encoding):
        conn = self._get_db()
        escape = self._escape_args
        if isinstance(prefix, str):
            prefix = prefix.encode(encoding)
        if isinstance(postfix, str):
            postfix = postfix.encode(encoding)
        sql = bytearray(prefix)
        args = iter(args)
        v = values % escape(next(args), conn)
        if isinstance(v, str):
            v = v.encode(encoding, 'surrogateescape')
        sql += v
        rows = 0
        for arg in args:
            v = values % escape(arg, conn)
            if isinstance(v, str):
                v = v.encode(encoding, 'surrogateescape')
            if len(sql) + len(v) + len(postfix) + 1 > max_stmt_length:
                r = await self.execute(sql + postfix)
                rows += r
                sql = bytearray(prefix)
            else:
                sql += b','
            sql += v
        r = await self.execute(sql + postfix)
        rows += r
        self._rowcount = rows
        return rows

    async def callproc(self, procname, args=()):
        """Execute stored procedure procname with args

        Compatibility warning: PEP-249 specifies that any modified
        parameters must be returned. This is currently impossible
        as they are only available by storing them in a server
        variable and then retrieved by a query. Since stored
        procedures return zero or more result sets, there is no
        reliable way to get at OUT or INOUT parameters via callproc.
        The server variables are named @_procname_n, where procname
        is the parameter above and n is the position of the parameter
        (from zero). Once all result sets generated by the procedure
        have been fetched, you can issue a SELECT @_procname_0, ...
        query using .execute() to get any OUT or INOUT values.

        Compatibility warning: The act of calling a stored procedure
        itself creates an empty result set. This appears after any
        result sets generated by the procedure. This is non-standard
        behavior with respect to the DB-API. Be sure to use nextset()
        to advance through all result sets; otherwise you may get
        disconnected.

        :param procname: ``str``, name of procedure to execute on server
        :param args: `sequence of parameters to use with procedure
        :returns: the original args.
        """
        conn = self._get_db()
        if self._echo:
            logger.info("CALL %s", procname)
            logger.info("%r", args)

        for index, arg in enumerate(args):
            q = "SET @_%s_%d=%s" % (procname, index, conn.escape(arg))
            await self._query(q)
            await self.nextset()

        _args = ','.join('@_%s_%d' % (procname, i) for i in range(len(args)))
        q = "CALL %s(%s)" % (procname, _args)
        await self._query(q)
        self._executed = q
        return args

    def fetchone(self):
        """Fetch the next row """
        self._check_executed()
        fut = self._loop.create_future()

        if self._rows is None or self._rownumber >= len(self._rows):
            fut.set_result(None)
            return fut
        result = self._rows[self._rownumber]
        self._rownumber += 1

        fut = self._loop.create_future()
        fut.set_result(result)
        return fut

    def fetchmany(self, size=None):
        """Returns the next set of rows of a query result, returning a
        list of tuples. When no more rows are available, it returns an
        empty list.

        The number of rows returned can be specified using the size argument,
        which defaults to one

        :param size: ``int`` number of rows to return
        :returns: ``list`` of fetched rows
        """
        self._check_executed()
        fut = self._loop.create_future()
        if self._rows is None:
            fut.set_result([])
            return fut
        end = self._rownumber + (size or self._arraysize)
        result = self._rows[self._rownumber:end]
        self._rownumber = min(end, len(self._rows))

        fut.set_result(result)
        return fut

    def fetchall(self):
        """Returns all rows of a query result set

        :returns: ``list`` of fetched rows
        """
        self._check_executed()
        fut = self._loop.create_future()
        if self._rows is None:
            fut.set_result([])
            return fut

        if self._rownumber:
            result = self._rows[self._rownumber:]
        else:
            result = self._rows
        self._rownumber = len(self._rows)

        fut.set_result(result)
        return fut

    def scroll(self, value, mode='relative'):
        """Scroll the cursor in the result set to a new position according
         to mode.

        If mode is relative (default), value is taken as offset to the
        current position in the result set, if set to absolute, value
        states an absolute target position. An IndexError should be raised in
        case a scroll operation would leave the result set. In this case,
        the cursor position is left undefined (ideal would be to
        not move the cursor at all).

        :param int value: move cursor to next position according to mode.
        :param str mode: scroll mode, possible modes: `relative` and `absolute`
        """
        self._check_executed()
        if mode == 'relative':
            r = self._rownumber + value
        elif mode == 'absolute':
            r = value
        else:
            raise ProgrammingError("unknown scroll mode %s" % mode)

        if not (0 <= r < len(self._rows)):
            raise IndexError("out of range")
        self._rownumber = r

        fut = self._loop.create_future()
        fut.set_result(None)
        return fut

    async def _query(self, q):
        conn = self._get_db()
        self._last_executed = q
        self._clear_result()
        await conn.query(q)
        await self._do_get_result()

    def _clear_result(self):
        self._rownumber = 0
        self._result = None

        self._rowcount = 0
        self._description = None
        self._lastrowid = None
        self._rows = None

    async def _do_get_result(self):
        conn = self._get_db()
        self._rownumber = 0
        self._result = result = conn._result
        self._rowcount = result.affected_rows
        self._description = result.description
        self._lastrowid = result.insert_id
        self._rows = result.rows

        if result.warning_count > 0:
            await self._show_warnings(conn)

    async def _show_warnings(self, conn):
        if self._result and self._result.has_next:
            return
        ws = await conn.show_warnings()
        if ws is None:
            return
        for w in ws:
            msg = w[-1]
            warnings.warn(str(msg), Warning, 4)

    Warning = Warning
    Error = Error
    InterfaceError = InterfaceError
    DatabaseError = DatabaseError
    DataError = DataError
    OperationalError = OperationalError
    IntegrityError = IntegrityError
    InternalError = InternalError
    ProgrammingError = ProgrammingError
    NotSupportedError = NotSupportedError

    def __aiter__(self):
        return self

    async def __anext__(self):
        ret = await self.fetchone()
        if ret is not None:
            return ret
        else:
            raise StopAsyncIteration  # noqa

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        return


class _DeserializationCursorMixin:
    async def _do_get_result(self):
        await super()._do_get_result()
        if self._rows:
            self._rows = [self._deserialization_row(r) for r in self._rows]

    def _deserialization_row(self, row):
        if row is None:
            return None
        if isinstance(row, dict):
            dict_flag = True
        else:
            row = list(row)
            dict_flag = False
        for index, (name, field_type, *n) in enumerate(self._description):
            if field_type == FIELD_TYPE.JSON:
                point = name if dict_flag else index
                with contextlib.suppress(ValueError, TypeError):
                    row[point] = json.loads(row[point])
        if dict_flag:
            return row
        else:
            return tuple(row)

    def _conv_row(self, row):
        if row is None:
            return None
        row = super()._conv_row(row)
        return self._deserialization_row(row)


class DeserializationCursor(_DeserializationCursorMixin, Cursor):
    """A cursor automatic deserialization of json type fields"""


class _DictCursorMixin:
    # You can override this to use OrderedDict or other dict-like types.
    dict_type = dict

    async def _do_get_result(self):
        await super()._do_get_result()
        fields = []
        if self._description:
            for f in self._result.fields:
                name = f.name
                if name in fields:
                    name = f.table_name + '.' + name
                fields.append(name)
            self._fields = fields

        if fields and self._rows:
            self._rows = [self._conv_row(r) for r in self._rows]

    def _conv_row(self, row):
        if row is None:
            return None
        row = super()._conv_row(row)
        return self.dict_type(zip(self._fields, row))


class DictCursor(_DictCursorMixin, Cursor):
    """A cursor which returns results as a dictionary"""


class SSCursor(Cursor):
    """Unbuffered Cursor, mainly useful for queries that return a lot of
    data, or for connections to remote servers over a slow network.

    Instead of copying every row of data into a buffer, this will fetch
    rows as needed. The upside of this, is the client uses much less memory,
    and rows are returned much faster when traveling over a slow network,
    or if the result set is very big.

    There are limitations, though. The MySQL protocol doesn't support
    returning the total number of rows, so the only way to tell how many rows
    there are is to iterate over every row returned. Also, it currently isn't
    possible to scroll backwards, as only the current row is held in memory.
    """

    async def close(self):
        conn = self._connection
        if conn is None:
            return

        if self._result is not None and self._result is conn._result:
            await self._result._finish_unbuffered_query()

        try:
            while (await self.nextset()):
                pass
        finally:
            self._connection = None

    async def _query(self, q):
        conn = self._get_db()
        self._last_executed = q
        await conn.query(q, unbuffered=True)
        await self._do_get_result()
        return self._rowcount

    async def _read_next(self):
        """Read next row """
        row = await self._result._read_rowdata_packet_unbuffered()
        row = self._conv_row(row)
        return row

    async def fetchone(self):
        """ Fetch next row """
        self._check_executed()
        row = await self._read_next()
        if row is None:
            return
        self._rownumber += 1
        return row

    async def fetchall(self):
        """Fetch all, as per MySQLdb. Pretty useless for large queries, as
        it is buffered.
        """
        rows = []
        while True:
            row = await self.fetchone()
            if row is None:
                break
            rows.append(row)
        return rows

    async def fetchmany(self, size=None):
        """Returns the next set of rows of a query result, returning a
        list of tuples. When no more rows are available, it returns an
        empty list.

        The number of rows returned can be specified using the size argument,
        which defaults to one

        :param size: ``int`` number of rows to return
        :returns: ``list`` of fetched rows
        """
        self._check_executed()
        if size is None:
            size = self._arraysize

        rows = []
        for i in range(size):
            row = await self._read_next()
            if row is None:
                break
            rows.append(row)
            self._rownumber += 1
        return rows

    async def scroll(self, value, mode='relative'):
        """Scroll the cursor in the result set to a new position
        according to mode . Same as :meth:`Cursor.scroll`, but move cursor
        on server side one by one row. If you want to move 20 rows forward
        scroll will make 20 queries to move cursor. Currently only forward
        scrolling is supported.

        :param int value: move cursor to next position according to mode.
        :param str mode: scroll mode, possible modes: `relative` and `absolute`
        """

        self._check_executed()

        if mode == 'relative':
            if value < 0:
                raise NotSupportedError("Backwards scrolling not supported "
                                        "by this cursor")

            for _ in range(value):
                await self._read_next()
            self._rownumber += value
        elif mode == 'absolute':
            if value < self._rownumber:
                raise NotSupportedError(
                    "Backwards scrolling not supported by this cursor")

            end = value - self._rownumber
            for _ in range(end):
                await self._read_next()
            self._rownumber = value
        else:
            raise ProgrammingError("unknown scroll mode %s" % mode)


class SSDictCursor(_DictCursorMixin, SSCursor):
    """An unbuffered cursor, which returns results as a dictionary """



from enum import IntEnum
from unittest import mock

import pytest
from sqlalchemy import MetaData, Table, Column, Integer, TypeDecorator

from aiomysql import sa


class UserDefinedEnum(IntEnum):
    Value1 = 111
    Value2 = 222


class IntEnumField(TypeDecorator):
    impl = Integer

    def __init__(self, enum_class, *arg, **kw):
        TypeDecorator.__init__(self, *arg, **kw)
        self.enum_class = enum_class

    def process_bind_param(self, value, dialect):
        """ From python to DB """
        if value is None:
            return None
        elif not isinstance(value, self.enum_class):
            return self.enum_class(value).value
        else:
            return value.value

    def process_result_value(self, value, dialect):
        """ From DB to Python """
        if value is None:
            return None

        return self.enum_class(value)


meta = MetaData()
tbl = Table('sa_test_type_tbl', meta,
            Column('id', Integer, nullable=False,
                   primary_key=True),
            Column('val', IntEnumField(enum_class=UserDefinedEnum)))


@pytest.fixture()
def sa_connect(connection_creator):
    async def connect(**kwargs):
        conn = await connection_creator()
        await conn.autocommit(True)
        cur = await conn.cursor()
        await cur.execute("DROP TABLE IF EXISTS sa_test_type_tbl")
        await cur.execute("CREATE TABLE sa_test_type_tbl "
                          "(id serial, val bigint)")
        await cur._connection.commit()
        engine = mock.Mock()
        engine.dialect = sa.engine._dialect
        return sa.SAConnection(conn, engine)
    return connect


@pytest.mark.run_loop
async def test_values(sa_connect):
    conn = await sa_connect()

    await conn.execute(tbl.insert().values(
        val=UserDefinedEnum.Value1)
    )
    result = await conn.execute(tbl.select().where(
        tbl.c.val == UserDefinedEnum.Value1)
    )
    data = await result.fetchone()
    assert data['val'] == UserDefinedEnum.Value1



import asyncio
import aiomysql


async def test_example_transaction(loop):
    conn = await aiomysql.connect(host='127.0.0.1', port=3306,
                                  user='root', password='',
                                  db='test_pymysql', autocommit=False,
                                  loop=loop)

    async with conn.cursor() as cursor:
        stmt_drop = "DROP TABLE IF EXISTS names"
        await cursor.execute(stmt_drop)
        await cursor.execute("""
            CREATE TABLE names (
            id TINYINT UNSIGNED NOT NULL AUTO_INCREMENT,
            name VARCHAR(30) DEFAULT '' NOT NULL,
            cnt TINYINT UNSIGNED DEFAULT 0,
            PRIMARY KEY (id))""")
        await conn.commit()

        # Insert 3 records
        names = (('Geert',), ('Jan',), ('Michel',))
        stmt_insert = "INSERT INTO names (name) VALUES (%s)"
        await cursor.executemany(stmt_insert, names)

        # Roll back!!!!
        await conn.rollback()

        # There should be no data!
        stmt_select = "SELECT id, name FROM names ORDER BY id"
        await cursor.execute(stmt_select)
        resp = await cursor.fetchall()
        # Check there is no data
        assert not resp

        # Do the insert again.
        await cursor.executemany(stmt_insert, names)

        # Data should be already there
        await cursor.execute(stmt_select)
        resp = await cursor.fetchall()
        print(resp)
        # Do a commit
        await conn.commit()

        await cursor.execute(stmt_select)
        print(resp)

        # Cleaning up, dropping the table again
        await cursor.execute(stmt_drop)
        await cursor.close()
        conn.close()


loop = asyncio.get_event_loop()
loop.run_until_complete(test_example_transaction(loop))



import pytest
from aiomysql import DictCursor


@pytest.fixture
def table(loop, connection, table_cleanup):
    async def f():
        cursor = await connection.cursor(DictCursor)
        sql = """CREATE TABLE bulkinsert (id INT(11), name CHAR(20),
                 age INT, height INT, PRIMARY KEY (id))"""
        await cursor.execute(sql)
    table_cleanup('bulkinsert')
    loop.run_until_complete(f())


@pytest.fixture
def assert_records(cursor):
    async def f(data):
        await cursor.execute(
            "SELECT id, name, age, height FROM bulkinsert")
        result = await cursor.fetchall()
        await cursor.execute('COMMIT')
        assert sorted(data) == sorted(result)
    return f


@pytest.fixture
def assert_dict_records(connection):
    async def f(data):
        cursor = await connection.cursor(DictCursor)
        await cursor.execute(
            "SELECT id, name, age, height FROM bulkinsert")
        result = await cursor.fetchall()
        await cursor.execute('COMMIT')
        assert sorted(data, key=lambda k: k['id']) == \
            sorted(result, key=lambda k: k['id'])
    return f


@pytest.mark.run_loop
async def test_bulk_insert(cursor, table, assert_records):
    data = [(0, "bob", 21, 123), (1, "jim", 56, 45), (2, "fred", 100, 180)]
    await cursor.executemany(
        "INSERT INTO bulkinsert (id, name, age, height) "
        "VALUES (%s,%s,%s,%s)", data)
    expected = bytearray(b"INSERT INTO bulkinsert (id, name, age, height) "
                         b"VALUES (0,'bob',21,123),(1,'jim',56,45),"
                         b"(2,'fred',100,180)")
    assert cursor._last_executed == expected
    await cursor.execute('commit')
    await assert_records(data)


@pytest.mark.run_loop
async def test_bulk_insert_multiline_statement(cursor, table, assert_records):
    data = [(0, "bob", 21, 123), (1, "jim", 56, 45), (2, "fred", 100, 180)]
    await cursor.executemany("""insert
        into bulkinsert (id, name,
        age, height)
        values (%s,
        %s , %s,
        %s )
         """, data)
    assert cursor._last_executed.strip() == bytearray(b"""insert
        into bulkinsert (id, name,
        age, height)
        values (0,
        'bob' , 21,
        123 ),(1,
        'jim' , 56,
        45 ),(2,
        'fred' , 100,
        180 )""")
    await cursor.execute('COMMIT')
    await assert_records(data)


@pytest.mark.run_loop
async def test_bulk_insert_single_record(cursor, table, assert_records):
    data = [(0, "bob", 21, 123)]
    await cursor.executemany(
        "insert into bulkinsert (id, name, age, height) "
        "values (%s,%s,%s,%s)", data)
    await cursor.execute('COMMIT')
    await assert_records(data)


@pytest.mark.run_loop
async def test_insert_on_duplicate_key_update(cursor, table, assert_records):
    # executemany should work with "insert ... on update" "
    data = [(0, "bob", 21, 123), (1, "jim", 56, 45), (2, "fred", 100, 180)]
    await cursor.executemany("""insert
        into bulkinsert (id, name,
        age, height)
        values (%s,
        %s , %s,
        %s ) on duplicate key update
        age = values(age)
         """, data)
    assert cursor._last_executed.strip() == bytearray(b"""insert
        into bulkinsert (id, name,
        age, height)
        values (0,
        'bob' , 21,
        123 ),(1,
        'jim' , 56,
        45 ),(2,
        'fred' , 100,
        180 ) on duplicate key update
        age = values(age)""")
    await cursor.execute('COMMIT')
    await assert_records(data)


@pytest.mark.run_loop
async def test_bulk_insert_with_params_as_dict(cursor, table,
                                               assert_dict_records):
    data = [
        {
            'id': 0,
            'name': "bob",
            'age': 21,
            'height': 123
        },
        {
            'id': 1,
            'name': "jim",
            'age': 56,
            'height': 45
        },
        {
            'id': 2,
            'name': "fred",
            'age': 100,
            'height': 180
        },
    ]
    await cursor.executemany(
        "INSERT INTO bulkinsert (id, name, age, height) "
        "VALUES (%(id)s,%(name)s,%(age)s,%(height)s)", data)
    expected = bytearray(b"INSERT INTO bulkinsert (id, name, age, height) "
                         b"VALUES (0,'bob',21,123),(1,'jim',56,45),"
                         b"(2,'fred',100,180)")
    assert cursor._last_executed == expected
    await cursor.execute('commit')
    await assert_dict_records(data)


@pytest.mark.run_loop
async def test_bulk_insert_with_precedence_spaces(cursor, table,
                                                  assert_records):
    data = [(0, "bob", 21, 123), (1, "jim", 56, 45)]
    await cursor.executemany("""
        INSERT INTO bulkinsert (id, name, age, height)
        VALUES (%s,%s,%s,%s)
    """, data)
    expected = bytearray(b"INSERT INTO bulkinsert (id, name, age, height)"
                         b"\n        VALUES (0,\'bob\',21,123),"
                         b"(1,\'jim\',56,45)\n    ")
    assert cursor._last_executed == expected
    await cursor.execute('commit')
    await assert_records(data)


@pytest.mark.run_loop
async def test_bulk_replace(cursor, table, assert_records):
    data = [(0, "bob", 21, 123), (0, "jim", 56, 45)]
    sql = ("REPLACE INTO bulkinsert (id, name, age, height) " +
           "VALUES (%s,%s,%s,%s)")
    await cursor.executemany(sql, data)
    assert cursor._last_executed.strip() == bytearray(
        b"REPLACE INTO bulkinsert (id, name, age, height) " +
        b"VALUES (0,'bob',21,123),(0,'jim',56,45)"
    )
    await cursor.execute('COMMIT')
    await assert_records([(0, "jim", 56, 45)])


@pytest.mark.run_loop
async def test_bulk_insert_with_semicolon_at_the_end(cursor, table,
                                                     assert_records):
    data = [(0, "bob", 21, 123), (1, "jim", 56, 45)]
    await cursor.executemany(
        "INSERT INTO bulkinsert (id, name, age, height) "
        "VALUES (%s,%s,%s,%s);", data)
    expected = bytearray(b"INSERT INTO bulkinsert (id, name, age, height) "
                         b"VALUES (0,'bob',21,123),(1,'jim',56,45)")
    assert cursor._last_executed == expected
    await cursor.execute('commit')
    await assert_records(data)



import datetime

import pytest

import aiomysql.cursors


BOB = {'name': 'bob', 'age': 21,
       'DOB': datetime.datetime(1990, 2, 6, 23, 4, 56)}
JIM = {'name': 'jim', 'age': 56,
       'DOB': datetime.datetime(1955, 5, 9, 13, 12, 45)}
FRED = {'name': 'fred', 'age': 100,
        'DOB': datetime.datetime(1911, 9, 12, 1, 1, 1)}

CURSOR_TYPE = aiomysql.cursors.DictCursor


async def prepare(conn):
    c = await conn.cursor(CURSOR_TYPE)

    # create a table ane some data to query
    await c.execute("drop table if exists dictcursor")
    await c.execute(
        """CREATE TABLE dictcursor (name char(20), age int ,
        DOB datetime)""")
    data = [("bob", 21, "1990-02-06 23:04:56"),
            ("jim", 56, "1955-05-09 13:12:45"),
            ("fred", 100, "1911-09-12 01:01:01")]
    await c.executemany("insert into dictcursor values "
                        "(%s,%s,%s)", data)


@pytest.mark.run_loop
async def test_dictcursor(connection):
    conn = connection
    await prepare(connection)

    bob, jim, fred = BOB.copy(), JIM.copy(), FRED.copy()
    # all assert test compare to the structure as would come
    # out from MySQLdb
    c = await conn.cursor(CURSOR_TYPE)

    # try an update which should return no rows
    await c.execute("update dictcursor set age=20 where name='bob'")
    bob['age'] = 20
    # pull back the single row dict for bob and check
    await c.execute("SELECT * from dictcursor where name='bob'")
    r = await c.fetchone()
    assert bob == r, "fetchone via DictCursor failed"
    # same again, but via fetchall => tuple)
    await c.execute("SELECT * from dictcursor where name='bob'")
    r = await c.fetchall()
    assert [bob] == r, \
        "fetch a 1 row result via fetchall failed via DictCursor"

    # get all 3 row via fetchall
    await c.execute("SELECT * from dictcursor")
    r = await c.fetchall()
    assert [bob, jim, fred] == r, "fetchall failed via DictCursor"

    # get all 2 row via fetchmany
    await c.execute("SELECT * from dictcursor")
    r = await c.fetchmany(2)
    assert [bob, jim] == r, "fetchmany failed via DictCursor"
    await c.execute('commit')


@pytest.mark.run_loop
async def test_custom_dict(connection):
    conn = connection
    await prepare(connection)

    class MyDict(dict):
        pass

    class MyDictCursor(CURSOR_TYPE):
        dict_type = MyDict

    keys = ['name', 'age', 'DOB']
    bob = MyDict([(k, BOB[k]) for k in keys])
    jim = MyDict([(k, JIM[k]) for k in keys])
    fred = MyDict([(k, FRED[k]) for k in keys])

    cur = await conn.cursor(MyDictCursor)
    await cur.execute("SELECT * FROM dictcursor WHERE name='bob'")
    r = await cur.fetchone()
    assert bob == r, "fetchone() returns MyDictCursor"

    await cur.execute("SELECT * FROM dictcursor")
    r = await cur.fetchall()
    assert [bob, jim, fred] == r, "fetchall failed via MyDictCursor"

    await cur.execute("SELECT * FROM dictcursor")
    r = await cur.fetchmany(2)
    assert [bob, jim] == r, "list failed via MyDictCursor"


@pytest.mark.run_loop
async def test_ssdictcursor(connection):
    conn = connection
    await prepare(connection)

    c = await conn.cursor(aiomysql.cursors.SSDictCursor)
    await c.execute("SELECT * from dictcursor where name='bob'")
    r = await c.fetchall()
    assert [BOB] == r,\
        "fetch a 1 row result via fetchall failed via DictCursor"



import asyncio
import aiomysql


async def test_example(loop):
    conn = await aiomysql.connect(host='127.0.0.1', port=3306,
                                  user='root', password='', db='mysql',
                                  loop=loop)
    sql = "SELECT 1 `id`, JSON_OBJECT('key1', 1, 'key2', 'abc') obj"
    async with conn.cursor(aiomysql.cursors.DeserializationCursor,
                           aiomysql.cursors.DictCursor) as cur:
        await cur.execute(sql)
        print(cur.description)
        r = await cur.fetchall()
        print(r)
    conn.close()


loop = asyncio.get_event_loop()
loop.run_until_complete(test_example(loop))



import os
import aiomysql
from tests._testutils import BaseTest


class AIOPyMySQLTestCase(BaseTest):

    async def _connect_all(self):
        conn1 = await aiomysql.connect(loop=self.loop, host=self.host,
                                       port=self.port, user=self.user,
                                       db=self.db,
                                       password=self.password,
                                       use_unicode=True, echo=True)
        conn2 = await aiomysql.connect(loop=self.loop, host=self.host,
                                       port=self.port, user=self.user,
                                       db=self.other_db,
                                       password=self.password,
                                       use_unicode=False, echo=True)
        conn3 = await aiomysql.connect(loop=self.loop, host=self.host,
                                       port=self.port, user=self.user,
                                       db=self.db,
                                       password=self.password,
                                       use_unicode=True, echo=True,
                                       local_infile=True)

        self.connections = [conn1, conn2, conn3]

    def setUp(self):
        super(AIOPyMySQLTestCase, self).setUp()
        self.host = os.environ.get('MYSQL_HOST', 'localhost')
        self.port = int(os.environ.get('MYSQL_PORT', 3306))
        self.user = os.environ.get('MYSQL_USER', 'root')
        self.db = os.environ.get('MYSQL_DB', 'test_pymysql')
        self.other_db = os.environ.get('OTHER_MYSQL_DB', 'test_pymysql2')
        self.password = os.environ.get('MYSQL_PASSWORD', '')

        self.connections = []
        self.loop.run_until_complete(self._connect_all())

    def tearDown(self):
        for connection in self.connections:
            self.loop.run_until_complete(connection.ensure_closed())
        self.doCleanups()
        super(AIOPyMySQLTestCase, self).tearDown()

    async def connect(self, host=None, user=None, password=None,
                      db=None, use_unicode=True, port=None,
                      **kwargs):
        if host is None:
            host = self.host
        if user is None:
            user = self.user
        if password is None:
            password = self.password
        if db is None:
            db = self.db
        if port is None:
            port = self.port
        conn = await aiomysql.connect(loop=self.loop, host=host,
                                      user=user, password=password,
                                      db=db, use_unicode=use_unicode,
                                      port=port,
                                      **kwargs)
        self.addCleanup(conn.close)
        return conn

    async def create_pool(self, host=None, user=None, password=None,
                          db=None, use_unicode=True,
                          port=None, **kwargs):
        if host is None:
            host = self.host
        if user is None:
            user = self.user
        if password is None:
            password = self.password
        if db is None:
            db = self.db
        if port is None:
            port = self.port
        pool = await aiomysql.create_pool(loop=self.loop, host=host,
                                          user=user, password=password,
                                          db=db, use_unicode=use_unicode,
                                          port=port,
                                          **kwargs)
        self.addCleanup(pool.close)
        return pool



import asyncio

import pytest

from aiomysql import ProgrammingError, Cursor, InterfaceError, OperationalError
from aiomysql.cursors import RE_INSERT_VALUES


async def _prepare(conn):
    cur = await conn.cursor()
    await cur.execute("DROP TABLE IF EXISTS tbl;")

    await cur.execute("""CREATE TABLE tbl (
             id MEDIUMINT NOT NULL AUTO_INCREMENT,
             name VARCHAR(255) NOT NULL,
             PRIMARY KEY (id));""")

    for i in [(1, 'a'), (2, 'b'), (3, 'c')]:
        await cur.execute("INSERT INTO tbl VALUES(%s, %s)", i)
    await cur.execute("DROP TABLE IF EXISTS tbl2")
    await cur.execute("""CREATE TABLE tbl2
                              (id int, name varchar(255))""")
    await conn.commit()


async def _prepare_procedure(conn):
    cur = await conn.cursor()
    await cur.execute("DROP PROCEDURE IF EXISTS myinc;")
    await cur.execute("""CREATE PROCEDURE myinc(p1 INT)
                           BEGIN
                               SELECT p1 + 1;
                           END
                           """)
    await conn.commit()


@pytest.mark.run_loop
async def test_description(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    cur = await conn.cursor()
    assert cur.description is None
    await cur.execute('SELECT * from tbl;')

    assert len(cur.description) == 2, \
        'cursor.description describes too many columns'

    assert len(cur.description[0]) == 7, \
        'cursor.description[x] tuples must have 7 elements'

    assert cur.description[0][0].lower() == 'id', \
        'cursor.description[x][0] must return column name'

    assert cur.description[1][0].lower() == 'name', \
        'cursor.description[x][0] must return column name'

    # Make sure self.description gets reset, cursor should be
    # set to None in case of none resulting queries like DDL
    await cur.execute('DROP TABLE IF EXISTS foobar;')
    assert cur.description is None


@pytest.mark.run_loop
async def test_cursor_properties(connection_creator):
    conn = await connection_creator()
    cur = await conn.cursor()
    assert cur.connection is conn
    cur.setinputsizes()
    cur.setoutputsizes()
    assert cur.echo == conn.echo


@pytest.mark.run_loop
async def test_scroll_relative(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    cur = await conn.cursor()
    await cur.execute('SELECT * FROM tbl;')
    await cur.scroll(1)
    ret = await cur.fetchone()
    assert (2, 'b') == ret


@pytest.mark.run_loop
async def test_scroll_absolute(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    cur = await conn.cursor()
    await cur.execute('SELECT * FROM tbl;')
    await cur.scroll(2, mode='absolute')
    ret = await cur.fetchone()
    assert (3, 'c') == ret


@pytest.mark.run_loop
async def test_scroll_errors(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    cur = await conn.cursor()

    with pytest.raises(ProgrammingError):
        await cur.scroll(2, mode='absolute')

    cur = await conn.cursor()
    await cur.execute('SELECT * FROM tbl;')

    with pytest.raises(ProgrammingError):
        await cur.scroll(2, mode='not_valid_mode')


@pytest.mark.run_loop
async def test_scroll_index_error(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    cur = await conn.cursor()
    await cur.execute('SELECT * FROM tbl;')
    with pytest.raises(IndexError):
        await cur.scroll(1000)


@pytest.mark.run_loop
async def test_close(connection_creator):
    conn = await connection_creator()
    cur = await conn.cursor()
    await cur.close()
    assert cur.closed is True
    with pytest.raises(ProgrammingError):
        await cur.execute('SELECT 1')
    # try to close for second time
    await cur.close()


@pytest.mark.run_loop
async def test_arraysize(connection_creator):
    conn = await connection_creator()
    cur = await conn.cursor()
    assert 1 == cur.arraysize
    cur.arraysize = 10
    assert 10 == cur.arraysize


@pytest.mark.run_loop
async def test_rows(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)

    cur = await conn.cursor()
    await cur.execute('SELECT * from tbl')
    assert 3 == cur.rowcount
    assert 0 == cur.rownumber
    await cur.fetchone()
    assert 1 == cur.rownumber
    assert cur.lastrowid is None
    await cur.execute('INSERT INTO tbl VALUES (%s, %s)', (4, 'd'))
    assert 0 != cur.lastrowid
    await conn.commit()


@pytest.mark.run_loop
async def test_callproc(connection_creator):
    conn = await connection_creator()
    await _prepare_procedure(conn)
    cur = await conn.cursor()
    await cur.callproc('myinc', [1])
    ret = await cur.fetchone()
    assert (2,) == ret
    await cur.close()
    with pytest.raises(ProgrammingError):
        await cur.callproc('myinc', [1])
    conn.close()


@pytest.mark.run_loop
async def test_fetchone_no_result(connection_creator):
    # test a fetchone() with no rows
    conn = await connection_creator()
    c = await conn.cursor()
    await c.execute("create table test_nr (b varchar(32))")
    try:
        data = "pymysql"
        await c.execute("insert into test_nr (b) values (%s)", (data,))
        r = await c.fetchone()
        assert r is None
    finally:
        await c.execute("drop table test_nr")


@pytest.mark.run_loop
async def test_fetchmany_no_result(connection_creator):
    conn = await connection_creator()
    cur = await conn.cursor()
    await cur.execute('DROP TABLE IF EXISTS foobar;')
    r = await cur.fetchmany()
    assert [] == r


@pytest.mark.run_loop
async def test_fetchall_no_result(connection_creator):
    # test a fetchone() with no rows
    conn = await connection_creator()
    cur = await conn.cursor()
    await cur.execute('DROP TABLE IF EXISTS foobar;')
    r = await cur.fetchall()
    assert [] == r


@pytest.mark.run_loop
async def test_fetchall_with_scroll(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    cur = await conn.cursor()
    await cur.execute('SELECT * FROM tbl;')
    await cur.scroll(1)
    ret = await cur.fetchall()
    assert ((2, 'b'), (3, 'c')) == ret


@pytest.mark.run_loop
async def test_aggregates(connection_creator):
    """ test aggregate functions """
    conn = await connection_creator()
    c = await conn.cursor()
    try:
        await c.execute('create table test_aggregates (i integer)')
        for i in range(0, 10):
            await c.execute(
                'insert into test_aggregates (i) values (%s)', (i,))
        await c.execute('select sum(i) from test_aggregates')
        r, = await c.fetchone()
        assert sum(range(0, 10)) == r
    finally:
        await c.execute('drop table test_aggregates')


@pytest.mark.run_loop
async def test_single_tuple(connection_creator):
    """ test a single tuple """
    conn = await connection_creator()
    c = await conn.cursor()
    try:
        await c.execute(
            "create table mystuff (id integer primary key)")
        await c.execute("insert into mystuff (id) values (1)")
        await c.execute("insert into mystuff (id) values (2)")
        await c.execute("select id from mystuff where id in %s", ((1,),))
        r = await c.fetchall()
        assert [(1,)] == list(r)
    finally:
        await c.execute("drop table mystuff")


@pytest.mark.run_loop
async def test_executemany(connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    cur = await conn.cursor()
    assert cur.description is None
    args = [1, 2, 3]
    row_count = await cur.executemany(
        'SELECT * FROM tbl WHERE id  = %s;', args)
    assert row_count == 3
    r = await cur.fetchall()
    # TODO: if this right behaviour
    assert ((3, 'c'),) == r

    # calling execute many without args
    row_count = await cur.executemany('SELECT 1;', ())
    assert row_count is None


@pytest.mark.run_loop
async def test_custom_cursor(connection_creator):
    class MyCursor(Cursor):
        pass
    conn = await connection_creator()
    cur = await conn.cursor(MyCursor)
    assert isinstance(cur, MyCursor)
    await cur.execute("SELECT 42;")
    (r, ) = await cur.fetchone()
    assert r == 42


@pytest.mark.run_loop
async def test_custom_cursor_not_cursor_subclass(connection_creator):
    class MyCursor2:
        pass
    conn = await connection_creator()
    with pytest.raises(TypeError):
        await conn.cursor(MyCursor2)


@pytest.mark.run_loop
async def test_morgify(connection_creator):
    conn = await connection_creator()
    cur = await conn.cursor()
    pairs = [(1, 'a'), (2, 'b'), (3, 'c')]
    sql = "INSERT INTO tbl VALUES(%s, %s)"
    results = [cur.mogrify(sql, p) for p in pairs]
    expected = ["INSERT INTO tbl VALUES(1, 'a')",
                "INSERT INTO tbl VALUES(2, 'b')",
                "INSERT INTO tbl VALUES(3, 'c')"]
    assert results == expected


@pytest.mark.run_loop
async def test_execute_cancel(connection_creator):
    conn = await connection_creator()
    cur = await conn.cursor()
    # Cancel a cursor in the middle of execution, before it could
    # read even the first packet (SLEEP assures the timings)
    task = asyncio.ensure_future(cur.execute(
        "SELECT 1 as id, SLEEP(0.1) as xxx"))
    await asyncio.sleep(0.05)
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass

    with pytest.raises(InterfaceError):
        await conn.cursor()


@pytest.mark.run_loop
async def test_execute_percentage(connection_creator):
    # %% in column set
    conn = await connection_creator()
    async with conn.cursor() as cur:
        await cur.execute("DROP TABLE IF EXISTS percent_test")
        await cur.execute("""\
            CREATE TABLE percent_test (
                `A%` INTEGER,
                `B%` INTEGER)""")

        q = "INSERT INTO percent_test (`A%%`, `B%%`) VALUES (%s, %s)"

        await cur.execute(q, (3, 4))


@pytest.mark.run_loop
async def test_executemany_percentage(connection_creator):
    # %% in column set
    conn = await connection_creator()
    async with conn.cursor() as cur:
        await cur.execute("DROP TABLE IF EXISTS percent_test")
        await cur.execute("""\
            CREATE TABLE percent_test (
                `A%` INTEGER,
                `B%` INTEGER)""")

        q = "INSERT INTO percent_test (`A%%`, `B%%`) VALUES (%s, %s)"

        assert RE_INSERT_VALUES.match(q) is not None
        await cur.executemany(q, [(3, 4), (5, 6)])
        assert cur._last_executed.endswith(b"(3, 4),(5, 6)"), \
            "executemany with %% not in one query"


@pytest.mark.run_loop
async def test_max_execution_time(mysql_server, connection_creator):
    conn = await connection_creator()
    await _prepare(conn)
    async with conn.cursor() as cur:
        # MySQL MAX_EXECUTION_TIME takes ms
        # MariaDB max_statement_time takes seconds as int/float, introduced in 10.1

        # this will sleep 0.01 seconds per row
        if mysql_server["db_type"] == "mysql":
            sql = """
                  SELECT /*+ MAX_EXECUTION_TIME(2000) */
                  name, sleep(0.01) FROM tbl
                  """
        else:
            sql = """
                  SET STATEMENT max_statement_time=2 FOR
                  SELECT name, sleep(0.01) FROM tbl
                  """

        await cur.execute(sql)
        # unlike SSCursor, Cursor returns a tuple of tuples here
        assert (await cur.fetchall()) == (
            ("a", 0),
            ("b", 0),
            ("c", 0),
        )

        if mysql_server["db_type"] == "mysql":
            sql = """
                  SELECT /*+ MAX_EXECUTION_TIME(2000) */
                  name, sleep(0.01) FROM tbl
                  """
        else:
            sql = """
                  SET STATEMENT max_statement_time=2 FOR
                  SELECT name, sleep(0.01) FROM tbl
                  """
        await cur.execute(sql)
        assert (await cur.fetchone()) == ("a", 0)

        # this discards the previous unfinished query
        await cur.execute("SELECT 1")
        assert (await cur.fetchone()) == (1,)

        if mysql_server["db_type"] == "mysql":
            sql = """
                  SELECT /*+ MAX_EXECUTION_TIME(1) */
                  name, sleep(1) FROM tbl
                  """
        else:
            sql = """
                  SET STATEMENT max_statement_time=0.001 FOR
                  SELECT name, sleep(1) FROM tbl
                  """
        with pytest.raises(OperationalError) as cm:
            # in a buffered cursor this should reliably raise an
            # OperationalError
            await cur.execute(sql)

        if mysql_server["db_type"] == "mysql":
            # this constant was only introduced in MySQL 5.7, not sure
            # what was returned before, may have been ER_QUERY_INTERRUPTED

            # this constant is pending a new PyMySQL release
            # assert cm.value.args[0] == pymysql.constants.ER.QUERY_TIMEOUT
            assert cm.value.args[0] == 3024
        else:
            # this constant is pending a new PyMySQL release
            # assert cm.value.args[0] == pymysql.constants.ER.STATEMENT_TIMEOUT
            assert cm.value.args[0] == 1969

        # connection should still be fine at this point
        await cur.execute("SELECT 1")
        assert (await cur.fetchone()) == (1,)



import warnings

import aiomysql
import pytest

from aiomysql import sa, create_pool, DictCursor, Cursor
from sqlalchemy import MetaData, Table, Column, Integer, String


meta = MetaData()
tbl = Table('tbl', meta,
            Column('id', Integer, nullable=False, primary_key=True),
            Column('name', String(255)))


@pytest.fixture
def table(loop, connection_creator, table_cleanup):
    async def f():
        connection = await connection_creator()
        cursor = await connection.cursor()
        await cursor.execute("DROP TABLE IF EXISTS tbl;")
        await cursor.execute("""CREATE TABLE tbl (
                 id MEDIUMINT NOT NULL AUTO_INCREMENT,
                 name VARCHAR(255) NOT NULL,
                 PRIMARY KEY (id));""")

        for i in [(1, 'a'), (2, 'b'), (3, 'c')]:
            await cursor.execute("INSERT INTO tbl VALUES(%s, %s)", i)

        await cursor.execute("commit;")
        await cursor.close()

    table_cleanup('tbl')
    loop.run_until_complete(f())


@pytest.mark.run_loop
async def test_cursor(table, cursor):
    ret = []
    await cursor.execute('SELECT * from tbl;')

    assert not cursor.closed
    async with cursor:
        async for i in cursor:
            ret.append(i)

    assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret
    assert cursor.closed


@pytest.mark.run_loop
async def test_cursor_lightweight(table, cursor):
    await cursor.execute('SELECT * from tbl;')

    assert not cursor.closed
    async with cursor:
        pass

    assert cursor.closed


@pytest.mark.run_loop
async def test_cursor_method(connection):
    async with connection.cursor() as cursor:
        await cursor.execute('SELECT 42;')
        value = await cursor.fetchone()
        assert value == (42,)

    assert cursor.closed


@pytest.mark.run_loop
async def test_connection(connection):
    assert not connection.closed
    async with connection:
        assert not connection.closed

    assert connection.closed


@pytest.mark.run_loop
async def test_connection_exception(connection):
    assert not connection.closed
    with pytest.raises(RuntimeError) as ctx:
        async with connection:
            assert not connection.closed
            raise RuntimeError('boom')
    assert str(ctx.value) == 'boom'
    assert connection.closed


@pytest.mark.run_loop
async def test_connect_method(mysql_params, loop):
    async with aiomysql.connect(loop=loop, **mysql_params) as connection:
        async with connection.cursor() as cursor:
            await cursor.execute("SELECT 42")
            value = await cursor.fetchone()
            assert value, (42,)

    assert cursor.closed
    assert connection.closed


@pytest.mark.run_loop
async def test_connect_method_exception(mysql_params, loop):
    with pytest.raises(RuntimeError) as ctx:
        async with aiomysql.connect(loop=loop, **mysql_params) as connection:
            assert not connection.closed
            raise RuntimeError('boom')

    assert str(ctx.value) == 'boom'
    assert connection.closed


@pytest.mark.run_loop
async def test_pool(table, pool_creator, loop):
    pool = await pool_creator()
    async with pool.acquire() as conn:
        async with (await conn.cursor()) as cur:
            await cur.execute("SELECT * from tbl")
            ret = []
            async for i in cur:
                ret.append(i)
            assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret


@pytest.mark.run_loop
async def test_create_pool_deprecations(mysql_params, loop):
    async with create_pool(loop=loop, **mysql_params) as pool:
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            async with pool.get() as conn:
                pass
    # The first warning emitted is expected to be DeprecationWarning:
    # in the past, we used to check for the last one but this assumption
    # breaks under Python 3.7 that also emits a `ResourceWarning` when
    # executed with `PYTHONASYNCIODEBUG=1`.
    assert issubclass(w[0].category, DeprecationWarning)
    assert conn.closed

    async with create_pool(loop=loop, **mysql_params) as pool:
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with await pool as conn:
                pass
    assert issubclass(w[-1].category, DeprecationWarning)
    assert conn.closed


@pytest.mark.run_loop
async def test_sa_connection(table, mysql_params, loop):
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        connection = await engine.acquire()
        assert not connection.closed
        async with connection:
            async with connection.execute(tbl.select()) as cursor:
                ret = []
                async for i in cursor:
                    ret.append(i)
            assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret
        assert connection.closed


@pytest.mark.run_loop
async def test_sa_transaction(table, mysql_params, loop):
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        async with engine.acquire() as connection:
            cnt = await connection.scalar(tbl.count())
            assert 3 == cnt

            async with (await connection.begin()) as tr:
                assert tr.is_active
                await connection.execute(tbl.delete())

            assert not tr.is_active
            cnt = await connection.scalar(tbl.count())
            assert 0 == cnt


@pytest.mark.run_loop
async def test_sa_transaction_rollback(loop, mysql_params, table):
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        async with engine.acquire() as conn:
            cnt = await conn.scalar(tbl.count())
            assert 3 == cnt

            with pytest.raises(RuntimeError) as ctx:
                async with (await conn.begin()) as tr:
                    assert tr.is_active
                    await conn.execute(tbl.delete())
                    raise RuntimeError("Exit")
            assert str(ctx.value) == "Exit"
            assert not tr.is_active
            cnt = await conn.scalar(tbl.count())
            assert 3 == cnt


@pytest.mark.run_loop
async def test_create_engine(loop, mysql_params, table):
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        async with engine.acquire() as conn:
            async with conn.execute(tbl.select()) as cursor:
                ret = []
                async for i in cursor:
                    ret.append(i)
                assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret


@pytest.mark.run_loop
async def test_engine(loop, mysql_params, table):
    engine = await sa.create_engine(loop=loop, **mysql_params)
    async with engine:
        async with engine.acquire() as conn:
            async with conn.execute(tbl.select()) as cursor:
                ret = []
                async for i in cursor:
                    ret.append(i)
                assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret


@pytest.mark.run_loop
async def test_transaction_context_manager(loop, mysql_params, table):
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        async with engine.acquire() as conn:
            async with conn.begin() as tr:
                async with conn.execute(tbl.select()) as cursor:
                    ret = []
                    async for i in cursor:
                        ret.append(i)
                    assert [(1, 'a'), (2, 'b'), (3, 'c')] == ret
                assert cursor.closed
            assert not tr.is_active

            tr2 = await conn.begin()
            async with tr2:
                assert tr2.is_active
                async with conn.execute('SELECT 1;') as cursor:
                    rec = await cursor.scalar()
                    assert rec == 1
                    await cursor.close()
            assert not tr2.is_active


@pytest.mark.run_loop
async def test_transaction_context_manager_error(loop, mysql_params, table):
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        async with engine.acquire() as conn:
            with pytest.raises(RuntimeError) as ctx:
                async with conn.begin() as tr:
                    assert tr.is_active
                    raise RuntimeError('boom')
            assert str(ctx.value) == 'boom'
            assert not tr.is_active
    assert conn.closed


@pytest.mark.run_loop
async def test_transaction_context_manager_commit_once(loop, mysql_params,
                                                       table):
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        async with engine.acquire() as conn:
            async with conn.begin() as tr:
                # check that in context manager we do not execute
                # commit for second time. Two commits in row causes
                # InvalidRequestError exception
                await tr.commit()
            assert not tr.is_active

            tr2 = await conn.begin()
            async with tr2:
                assert tr2.is_active
                # check for double commit one more time
                await tr2.commit()
            assert not tr2.is_active
    assert conn.closed


@pytest.mark.run_loop
async def test_incompatible_cursor_fails(loop, mysql_params):
    mysql_params['cursorclass'] = DictCursor
    with pytest.raises(sa.ArgumentError) as ctx:
        await sa.create_engine(loop=loop, **mysql_params)

    msg = 'SQLAlchemy engine does not support this cursor class'
    assert str(ctx.value) == msg


@pytest.mark.run_loop
async def test_compatible_cursor_correct(loop, mysql_params):
    class SubCursor(Cursor):
        pass

    mysql_params['cursorclass'] = SubCursor
    async with sa.create_engine(loop=loop, **mysql_params) as engine:
        async with engine.acquire() as conn:
            # check not raise sa.ArgumentError exception
            pass
    assert conn.closed



import asyncio

import pytest
from sqlalchemy import MetaData, Table, Column, Integer, String

from aiomysql import sa

meta = MetaData()
tbl = Table('sa_tbl3', meta,
            Column('id', Integer, nullable=False,
                   primary_key=True),
            Column('name', String(255)))


@pytest.fixture()
def make_engine(connection, mysql_params):
    async def _make_engine(**kwargs):
        if "unix_socket" in mysql_params:
            conn_args = {"unix_socket": mysql_params["unix_socket"]}
        else:
            conn_args = {
                "host": mysql_params['host'],
                "port": mysql_params['port'],
            }

        return (await sa.create_engine(db=mysql_params['db'],
                                       user=mysql_params['user'],
                                       password=mysql_params['password'],
                                       minsize=10,
                                       **conn_args,
                                       **kwargs))
    return _make_engine


async def start(engine):
    async with engine.acquire() as conn:
        await conn.execute("DROP TABLE IF EXISTS sa_tbl3")
        await conn.execute("CREATE TABLE sa_tbl3 "
                           "(id serial, name varchar(255))")


@pytest.mark.run_loop
async def test_dialect(make_engine):
    engine = await make_engine()
    await start(engine)

    assert sa.engine._dialect == engine.dialect


@pytest.mark.run_loop
async def test_name(make_engine):
    engine = await make_engine()
    await start(engine)
    assert 'mysql' == engine.name


@pytest.mark.run_loop
async def test_driver(make_engine):
    engine = await make_engine()
    await start(engine)
    assert 'pymysql' == engine.driver

# @pytest.mark.run_loop
# async def test_dsn(self):
#     self.assertEqual(
#         'dbname=aiomysql user=aiomysql password=xxxxxx host=127.0.0.1',
#         engine.dsn)


@pytest.mark.run_loop
async def test_minsize(make_engine):
    engine = await make_engine()
    await start(engine)
    assert 10 == engine.minsize


@pytest.mark.run_loop
async def test_maxsize(make_engine):
    engine = await make_engine()
    await start(engine)
    assert 10 == engine.maxsize


@pytest.mark.run_loop
async def test_size(make_engine):
    engine = await make_engine()
    await start(engine)
    assert 10 == engine.size


@pytest.mark.run_loop
async def test_freesize(make_engine):
    engine = await make_engine()
    await start(engine)
    assert 10 == engine.freesize


@pytest.mark.run_loop
async def test_make_engine_with_default_loop(make_engine):
    engine = await make_engine()
    await start(engine)

    engine.close()
    await engine.wait_closed()


@pytest.mark.run_loop
async def test_not_context_manager(make_engine):
    engine = await make_engine()
    await start(engine)
    with pytest.raises(RuntimeError):
        with engine:
            pass


@pytest.mark.run_loop
async def test_release_transacted(make_engine):
    engine = await make_engine()
    await start(engine)
    conn = await engine.acquire()
    tr = await conn.begin()
    with pytest.raises(sa.InvalidRequestError):
        engine.release(conn)
    del tr


@pytest.mark.run_loop
async def test_cannot_acquire_after_closing(make_engine):
    engine = await make_engine()
    await start(engine)
    engine.close()

    with pytest.raises(RuntimeError):
        await engine.acquire()
    await engine.wait_closed()


@pytest.mark.run_loop
async def test_wait_closed(make_engine):
    engine = await make_engine()
    await start(engine)

    c1 = await engine.acquire()
    c2 = await engine.acquire()
    assert 10 == engine.size
    assert 8 == engine.freesize

    ops = []

    async def do_release(conn):
        await asyncio.sleep(0)
        engine.release(conn)
        ops.append('release')

    async def wait_closed():
        await engine.wait_closed()
        ops.append('wait_closed')

    engine.close()
    await asyncio.gather(wait_closed(), do_release(c1),
                         do_release(c2))
    assert ['release', 'release', 'wait_closed'] == ops
    assert 0 == engine.freesize
    engine.close()
    await engine.wait_closed()


@pytest.mark.run_loop
async def test_terminate_with_acquired_connections(make_engine):
    engine = await make_engine()
    await start(engine)

    conn = await engine.acquire()
    engine.terminate()
    await engine.wait_closed()

    assert conn.closed



import asyncio
import aiomysql


async def test_example_executemany(loop):
    conn = await aiomysql.connect(host='127.0.0.1', port=3306,
                                       user='root', password='',
                                       db='test_pymysql', loop=loop)

    cur = await conn.cursor()
    async with conn.cursor() as cur:
        await cur.execute("DROP TABLE IF EXISTS music_style;")
        await cur.execute("""CREATE TABLE music_style
                                  (id INT,
                                  name VARCHAR(255),
                                  PRIMARY KEY (id));""")
        await conn.commit()

        # insert 3 rows one by one
        await cur.execute("INSERT INTO music_style VALUES(1,'heavy metal')")
        await cur.execute("INSERT INTO music_style VALUES(2,'death metal');")
        await cur.execute("INSERT INTO music_style VALUES(3,'power metal');")
        await conn.commit()

        # insert 3 row by one long query using *executemany* method
        data = [(4, 'gothic metal'), (5, 'doom metal'), (6, 'post metal')]
        await cur.executemany(
            "INSERT INTO music_style (id, name)"
            "values (%s,%s)", data)
        await conn.commit()

        # fetch all insert row from table music_style
        await cur.execute("SELECT * FROM music_style;")
        result = await cur.fetchall()
        print(result)

    conn.close()


loop = asyncio.get_event_loop()
loop.run_until_complete(test_example_executemany(loop))



import pytest
from sqlalchemy import bindparam
from sqlalchemy import MetaData, Table, Column, Integer, String

from aiomysql import sa


meta = MetaData()
tbl = Table('sa_tbl_cache_test', meta,
            Column('id', Integer, nullable=False,
                   primary_key=True),
            Column('val', String(255)))


@pytest.fixture()
def make_engine(mysql_params, connection):
    async def _make_engine(**kwargs):
        if "unix_socket" in mysql_params:
            conn_args = {"unix_socket": mysql_params["unix_socket"]}
        else:
            conn_args = {
                "host": mysql_params['host'],
                "port": mysql_params['port'],
            }

        return (await sa.create_engine(db=mysql_params['db'],
                                       user=mysql_params['user'],
                                       password=mysql_params['password'],
                                       minsize=10,
                                       **conn_args,
                                       **kwargs))

    return _make_engine


async def start(engine):
    async with engine.acquire() as conn:
        tx = await conn.begin()
        await conn.execute("DROP TABLE IF EXISTS "
                           "sa_tbl_cache_test")
        await conn.execute("CREATE TABLE sa_tbl_cache_test"
                           "(id serial, val varchar(255))")
        await conn.execute(tbl.insert().values(val='some_val_1'))
        await conn.execute(tbl.insert().values(val='some_val_2'))
        await conn.execute(tbl.insert().values(val='some_val_3'))
        await tx.commit()


@pytest.mark.run_loop
async def test_dialect(make_engine):
    cache = dict()
    engine = await make_engine(compiled_cache=cache)
    await start(engine)

    async with engine.acquire() as conn:
        # check select with params not added to cache
        q = tbl.select().where(tbl.c.val == 'some_val_1')
        cursor = await conn.execute(q)
        row = await cursor.fetchone()
        assert 'some_val_1' == row.val
        assert 0 == len(cache)

        # check select with bound params added to cache
        select_by_val = tbl.select().where(
            tbl.c.val == bindparam('value')
        )
        cursor = await conn.execute(
            select_by_val, {'value': 'some_val_3'}
        )
        row = await cursor.fetchone()
        assert 'some_val_3' == row.val
        assert 1 == len(cache)

        cursor = await conn.execute(
            select_by_val, value='some_val_2'
        )
        row = await cursor.fetchone()
        assert 'some_val_2' == row.val
        assert 1 == len(cache)

        select_all = tbl.select()
        cursor = await conn.execute(select_all)
        rows = await cursor.fetchall()
        assert 3 == len(rows)
        assert 2 == len(cache)

        # check insert with bound params not added to cache
        await conn.execute(tbl.insert().values(val='some_val_4'))
        assert 2 == len(cache)

        # check insert with bound params added to cache
        q = tbl.insert().values(val=bindparam('value'))
        await conn.execute(q, value='some_val_5')
        assert 3 == len(cache)

        await conn.execute(q, value='some_val_6')
        assert 3 == len(cache)

        await conn.execute(q, {'value': 'some_val_7'})
        assert 3 == len(cache)

        cursor = await conn.execute(select_all)
        rows = await cursor.fetchall()
        assert 7 == len(rows)
        assert 3 == len(cache)

        # check update with params not added to cache
        q = tbl.update().where(
            tbl.c.val == 'some_val_1'
        ).values(val='updated_val_1')
        await conn.execute(q)
        assert 3 == len(cache)
        cursor = await conn.execute(
            select_by_val, value='updated_val_1'
        )
        row = await cursor.fetchone()
        assert 'updated_val_1' == row.val

        # check update with bound params added to cache
        q = tbl.update().where(
            tbl.c.val == bindparam('value')
        ).values(val=bindparam('update'))
        await conn.execute(
            q, value='some_val_2', update='updated_val_2'
        )
        assert 4 == len(cache)
        cursor = await conn.execute(
            select_by_val, value='updated_val_2'
        )
        row = await cursor.fetchone()
        assert 'updated_val_2' == row.val



from aiomysql import create_pool

import pytest


@pytest.mark.run_loop
async def test_tls_connect(mysql_server, loop, mysql_params):
    if "unix_socket" in mysql_params:
        pytest.skip("TLS is not supported on unix sockets")

    async with create_pool(**mysql_server['conn_params'],
                           loop=loop) as pool:
        async with pool.get() as conn:
            async with conn.cursor() as cur:
                # Run simple command
                await cur.execute("SHOW DATABASES;")
                value = await cur.fetchall()

                values = [item[0] for item in value]
                # Spot check the answers, we should at least have mysql
                # and information_schema
                assert 'mysql' in values, \
                    'Could not find the "mysql" table'
                assert 'information_schema' in values, \
                    'Could not find the "mysql" table'

                # Check TLS variables
                await cur.execute("SHOW STATUS LIKE 'Ssl_version%';")
                value = await cur.fetchone()

                # The context has TLS
                assert value[1].startswith('TLS'), \
                    'Not connected to the database with TLS'


# MySQL will get you to renegotiate if sent a cleartext password
@pytest.mark.run_loop
async def test_auth_plugin_renegotiation(mysql_server, loop, mysql_params):
    if "unix_socket" in mysql_params:
        pytest.skip("TLS is not supported on unix sockets")

    async with create_pool(**mysql_server['conn_params'],
                           auth_plugin='mysql_clear_password',
                           loop=loop) as pool:
        async with pool.get() as conn:
            async with conn.cursor() as cur:
                # Run simple command
                await cur.execute("SHOW DATABASES;")
                value = await cur.fetchall()

                assert len(value), 'No databases found'

                # Check we tried to use the cleartext plugin
                assert conn._client_auth_plugin == 'mysql_clear_password', \
                    'Client did not try clear password auth'

                # Check the server asked us to use MySQL's default plugin
                assert conn._server_auth_plugin in (
                    'mysql_native_password', 'caching_sha2_password'), \
                    'Server did not ask for native auth'
                # Check we actually used the servers default plugin
                assert conn._auth_plugin_used in (
                    'mysql_native_password', 'caching_sha2_password'), \
                    'Client did not renegotiate with server\'s default auth'



import asyncio
import aiomysql


loop = asyncio.get_event_loop()


@asyncio.coroutine
def test_example():
    conn = yield from aiomysql.connect(
        host='127.0.0.1',
        port=3306,
        user='root',
        password='',
        db='mysql',
        loop=loop
    )

    cur = yield from conn.cursor()
    yield from cur.execute("SELECT Host,User FROM user")
    print(cur.description)
    r = yield from cur.fetchall()
    print(r)
    yield from cur.close()
    conn.close()


loop.run_until_complete(test_example())



import asyncio

import pytest
from aiomysql.connection import Connection
from aiomysql.pool import Pool


@pytest.mark.run_loop
async def test_create_pool(pool_creator):
    pool = await pool_creator()
    assert isinstance(pool, Pool)
    assert 1 == pool.minsize
    assert 10 == pool.maxsize
    assert 1 == pool.size
    assert 1 == pool.freesize
    assert not pool.echo


@pytest.mark.run_loop
async def test_create_pool2(pool_creator):
    pool = await pool_creator(minsize=10, maxsize=20)
    assert isinstance(pool, Pool)
    assert 10 == pool.minsize
    assert 20 == pool.maxsize
    assert 10 == pool.size
    assert 10 == pool.freesize


@pytest.mark.run_loop
async def test_acquire(pool_creator):
    pool = await pool_creator()
    conn = await pool.acquire()
    assert isinstance(conn, Connection)
    assert not conn.closed
    cursor = await conn.cursor()
    await cursor.execute('SELECT 1')
    val = await cursor.fetchone()
    await cursor.close()
    assert (1,) == val
    pool.release(conn)


@pytest.mark.run_loop
async def test_release(pool_creator):
    pool = await pool_creator()
    conn = await pool.acquire()
    assert 0 == pool.freesize
    assert {conn} == pool._used
    pool.release(conn)
    assert 1 == pool.freesize
    assert not pool._used


@pytest.mark.run_loop
async def test_release_closed(pool_creator):
    pool = await pool_creator(minsize=10, maxsize=10)
    conn = await pool.acquire()
    assert 9 == pool.freesize
    await conn.ensure_closed()
    pool.release(conn)
    assert 9 == pool.freesize
    assert not pool._used
    assert 9 == pool.size

    conn2 = await pool.acquire()
    assert 9 == pool.freesize
    assert 10 == pool.size
    pool.release(conn2)


@pytest.mark.run_loop
async def test_bad_context_manager_usage(pool_creator):
    pool = await pool_creator()
    with pytest.raises(RuntimeError):
        with pool:
            pass


@pytest.mark.run_loop
async def test_context_manager(pool_creator):
    pool = await pool_creator(minsize=10, maxsize=10)
    async with pool.get() as conn:
        assert isinstance(conn, Connection)
        assert 9 == pool.freesize
        assert {conn} == pool._used
    assert 10 == pool.freesize


@pytest.mark.run_loop
async def test_clear(pool_creator):
    pool = await pool_creator()
    await pool.clear()
    assert 0 == pool.freesize


@pytest.mark.run_loop
async def test_initial_empty(pool_creator):
    pool = await pool_creator(minsize=0)
    assert 10 == pool.maxsize
    assert 0 == pool.minsize
    assert 0 == pool.size
    assert 0 == pool.freesize

    async with pool.get():
        assert 1 == pool.size
        assert 0 == pool.freesize
    assert 1 == pool.size
    assert 1 == pool.freesize

    conn1 = await pool.acquire()
    assert 1 == pool.size
    assert 0 == pool.freesize

    conn2 = await pool.acquire()
    assert 2 == pool.size
    assert 0 == pool.freesize

    pool.release(conn1)
    assert 2 == pool.size
    assert 1 == pool.freesize

    pool.release(conn2)
    assert 2 == pool.size
    assert 2 == pool.freesize


@pytest.mark.run_loop
async def test_parallel_tasks(pool_creator, loop):
    pool = await pool_creator(minsize=0, maxsize=2)
    assert 2 == pool.maxsize
    assert 0 == pool.minsize
    assert 0 == pool.size
    assert 0 == pool.freesize

    fut1 = pool.acquire()
    fut2 = pool.acquire()

    conn1, conn2 = await asyncio.gather(fut1, fut2)
    assert 2 == pool.size
    assert 0 == pool.freesize
    assert {conn1, conn2} == pool._used

    pool.release(conn1)
    assert 2 == pool.size
    assert 1 == pool.freesize
    assert {conn2} == pool._used

    pool.release(conn2)
    assert 2 == pool.size
    assert 2 == pool.freesize
    assert not conn1.closed
    assert not conn2.closed

    conn3 = await pool.acquire()
    assert conn3 is conn1
    pool.release(conn3)


@pytest.mark.run_loop
async def test_parallel_tasks_more(pool_creator, loop):
    pool = await pool_creator(minsize=0, maxsize=3)

    fut1 = pool.acquire()
    fut2 = pool.acquire()
    fut3 = pool.acquire()

    conn1, conn2, conn3 = await asyncio.gather(fut1, fut2, fut3)
    assert 3 == pool.size
    assert 0 == pool.freesize
    assert {conn1, conn2, conn3} == pool._used

    pool.release(conn1)
    assert 3 == pool.size
    assert 1 == pool.freesize
    assert {conn2, conn3} == pool._used

    pool.release(conn2)
    assert 3 == pool.size
    assert 2 == pool.freesize
    assert {conn3} == pool._used
    assert not conn1.closed
    assert not conn2.closed

    pool.release(conn3)
    assert 3, pool.size
    assert 3, pool.freesize
    assert not pool._used
    assert not conn1.closed
    assert not conn2.closed
    assert not conn3.closed

    conn4 = await pool.acquire()
    assert conn4 is conn1
    pool.release(conn4)


@pytest.mark.run_loop
async def test_default_event_loop(pool_creator, loop):
    asyncio.set_event_loop(loop)
    pool = await pool_creator(loop=None)
    assert pool._loop is loop


@pytest.mark.run_loop
async def test_release_with_invalid_status(pool_creator):
    pool = await pool_creator(minsize=10, maxsize=10)
    conn = await pool.acquire()
    assert 9 == pool.freesize
    assert {conn} == pool._used
    cur = await conn.cursor()
    await cur.execute('BEGIN')
    await cur.close()

    pool.release(conn)
    assert 9 == pool.freesize
    assert not pool._used
    assert conn.closed


@pytest.mark.run_loop
async def test_release_with_invalid_status_wait_release(pool_creator):
    pool = await pool_creator(minsize=10, maxsize=10)
    conn = await pool.acquire()
    assert 9 == pool.freesize
    assert {conn} == pool._used
    cur = await conn.cursor()
    await cur.execute('BEGIN')
    await cur.close()

    await pool.release(conn)
    assert 9 == pool.freesize
    assert not pool._used
    assert conn.closed


@pytest.mark.run_loop
async def test__fill_free(pool_creator, loop):
    pool = await pool_creator(minsize=1)
    async with pool.get():
        assert 0 == pool.freesize
        assert 1 == pool.size

        conn = await asyncio.wait_for(pool.acquire(),
                                      timeout=0.5)
        assert 0 == pool.freesize
        assert 2 == pool.size
        pool.release(conn)
        assert 1 == pool.freesize
        assert 2 == pool.size
    assert 2 == pool.freesize
    assert 2 == pool.size


@pytest.mark.run_loop
async def test_connect_from_acquire(pool_creator):
    pool = await pool_creator(minsize=0)
    assert 0 == pool.freesize
    assert 0 == pool.size
    async with pool.get():
        assert 1 == pool.size
        assert 0 == pool.freesize
    assert 1 == pool.size
    assert 1 == pool.freesize


@pytest.mark.run_loop
async def test_concurrency(pool_creator):
    pool = await pool_creator(minsize=2, maxsize=4)
    c1 = await pool.acquire()
    c2 = await pool.acquire()
    assert 0 == pool.freesize
    assert 2 == pool.size
    pool.release(c1)
    pool.release(c2)


@pytest.mark.run_loop
async def test_invalid_minsize_and_maxsize(pool_creator):
    with pytest.raises(ValueError):
        await pool_creator(minsize=-1)

    with pytest.raises(ValueError):
        await pool_creator(minsize=5, maxsize=2)


@pytest.mark.run_loop
async def test_true_parallel_tasks(pool_creator, loop):
    pool = await pool_creator(minsize=0, maxsize=1)
    assert 1 == pool.maxsize
    assert 0 == pool.minsize
    assert 0 == pool.size
    assert 0 == pool.freesize

    maxsize = 0
    minfreesize = 100

    async def inner():
        nonlocal maxsize, minfreesize
        maxsize = max(maxsize, pool.size)
        minfreesize = min(minfreesize, pool.freesize)
        conn = await pool.acquire()
        maxsize = max(maxsize, pool.size)
        minfreesize = min(minfreesize, pool.freesize)
        await asyncio.sleep(0.01)
        pool.release(conn)
        maxsize = max(maxsize, pool.size)
        minfreesize = min(minfreesize, pool.freesize)

    await asyncio.gather(inner(), inner())

    assert 1 == maxsize
    assert 0 == minfreesize


@pytest.mark.run_loop
async def test_cannot_acquire_after_closing(pool_creator):
    pool = await pool_creator()
    pool.close()

    with pytest.raises(RuntimeError):
        await pool.acquire()


@pytest.mark.run_loop
async def test_wait_closed(pool_creator, loop):
    pool = await pool_creator(minsize=10, maxsize=10)

    c1 = await pool.acquire()
    c2 = await pool.acquire()
    assert 10 == pool.size
    assert 8 == pool.freesize

    ops = []

    async def do_release(conn):
        await asyncio.sleep(0)
        pool.release(conn)
        ops.append('release')

    async def wait_closed():
        await pool.wait_closed()
        ops.append('wait_closed')

    pool.close()
    await asyncio.gather(wait_closed(), do_release(c1), do_release(c2))
    assert ['release', 'release', 'wait_closed'] == ops
    assert 0 == pool.freesize
    assert pool.closed


@pytest.mark.run_loop
async def test_echo(pool_creator):
    pool = await pool_creator(echo=True)
    assert pool.echo

    async with pool.get() as conn:
        assert conn.echo


@pytest.mark.run_loop
async def test_terminate_with_acquired_connections(pool_creator):
    pool = await pool_creator()
    conn = await pool.acquire()
    pool.terminate()
    await pool.wait_closed()
    assert conn.closed


@pytest.mark.run_loop
async def test_release_closed_connection(pool_creator):
    pool = await pool_creator()
    conn = await pool.acquire()
    conn.close()

    pool.release(conn)
    pool.close()


@pytest.mark.run_loop
async def test_wait_closing_on_not_closed(pool_creator):
    pool = await pool_creator()

    with pytest.raises(RuntimeError):
        await pool.wait_closed()
    pool.close()


@pytest.mark.run_loop
async def test_release_terminated_pool(pool_creator):
    pool = await pool_creator()
    conn = await pool.acquire()
    pool.terminate()
    await pool.wait_closed()

    pool.release(conn)
    pool.close()


@pytest.mark.run_loop
async def test_release_terminated_pool_wait_release(pool_creator):
    pool = await pool_creator()
    conn = await pool.acquire()
    pool.terminate()
    await pool.wait_closed()

    await pool.release(conn)
    pool.close()


@pytest.mark.run_loop
async def test_close_with_acquired_connections(pool_creator, loop):
    pool = await pool_creator()
    conn = await pool.acquire()
    pool.close()

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(pool.wait_closed(), 0.1)
    pool.release(conn)


async def _set_global_conn_timeout(conn, t):
    # create separate connection to setup global connection timeouts
    # https://dev.mysql.com/doc/refman/5.1/en/server-system-variables
    # .html#sysvar_interactive_timeout
    cur = await conn.cursor()
    await cur.execute('SET GLOBAL wait_timeout=%s;', t)
    await cur.execute('SET GLOBAL interactive_timeout=%s;', t)
    await cur.close()


@pytest.mark.run_loop
async def test_drop_connection_if_timedout(pool_creator,
                                           connection_creator, loop):
    conn = await connection_creator()
    await _set_global_conn_timeout(conn, 2)
    await conn.ensure_closed()

    pool = conn = None
    try:
        pool = await pool_creator(minsize=3, maxsize=3)
        # sleep, more then connection timeout
        await asyncio.sleep(3)
        conn = await pool.acquire()
        cur = await conn.cursor()
        # query should not throw exception OperationalError
        await cur.execute('SELECT 1;')
        pool.release(conn)
        conn = None
        pool.close()
        await pool.wait_closed()
    finally:
        # TODO: this could probably be done better
        #       if this isn't closed it blocks forever
        try:
            if conn is not None:
                pool.release(conn)
            if pool is not None:
                pool.close()
                await pool.wait_closed()
        except Exception:
            pass

        # setup default timeouts
        conn = await connection_creator()
        await _set_global_conn_timeout(conn, 28800)
        await conn.ensure_closed()


@pytest.mark.skip(reason='Not implemented')
@pytest.mark.run_loop
async def test_create_pool_with_timeout(pool_creator):
    pool = await pool_creator(minsize=3, maxsize=3)
    timeout = 0.1
    assert timeout == pool.timeout
    conn = await pool.acquire()
    assert timeout == conn.timeout
    pool.release(conn)


@pytest.mark.run_loop
async def test_cancelled_connection(pool_creator, loop):
    pool = await pool_creator(minsize=0, maxsize=1)

    try:
        async with pool.get() as conn:
            curs = await conn.cursor()
            # Cancel a cursor in the middle of execution, before it
            # could read even the first packet (SLEEP assures the
            # timings)
            task = loop.create_task(curs.execute(
                "SELECT 1 as id, SLEEP(0.1) as xxx"))
            await asyncio.sleep(0.05)
            task.cancel()
            await task
    except asyncio.CancelledError:
        pass

    async with pool.get() as conn:
        cur2 = await conn.cursor()
        res = await cur2.execute("SELECT 2 as value, 0 as xxx")
        names = [x[0] for x in cur2.description]
        # If we receive ["id", "xxx"] - we corrupted the connection
        assert names == ["value", "xxx"]
        res = await cur2.fetchall()
        # If we receive [(1, 0)] - we retrieved old cursor's values
        assert list(res) == [(2, 0)]


@pytest.mark.run_loop
async def test_pool_with_connection_recycling(pool_creator, loop):
    pool = await pool_creator(minsize=1, maxsize=1, pool_recycle=3)
    async with pool.get() as conn:
        cur = await conn.cursor()
        await cur.execute('SELECT 1;')
        val = await cur.fetchone()
        assert (1,) == val

    await asyncio.sleep(5)

    assert 1 == pool.freesize
    async with pool.get() as conn:
        cur = await conn.cursor()
        await cur.execute('SELECT 1;')
        val = await cur.fetchone()
        assert (1,) == val


@pytest.mark.run_loop
async def test_pool_drops_connection_with_exception(pool_creator, loop):
    pool = await pool_creator(minsize=1, maxsize=1)

    async with pool.get() as conn:
        cur = await conn.cursor()
        await cur.execute('SELECT 1;')

    connection, = pool._free
    connection._writer._protocol.connection_lost(IOError())

    async with pool.get() as conn:
        cur = await conn.cursor()
        await cur.execute('SELECT 1;')


@pytest.mark.run_loop
async def test_pool_maxsize_unlimited(pool_creator, loop):
    pool = await pool_creator(minsize=0, maxsize=0)

    async with pool.acquire() as conn:
        cur = await conn.cursor()
        await cur.execute('SELECT 1;')


@pytest.mark.run_loop
async def test_pool_maxsize_unlimited_minsize_1(pool_creator, loop):
    pool = await pool_creator(minsize=1, maxsize=0)

    async with pool.acquire() as conn:
        cur = await conn.cursor()
        await cur.execute('SELECT 1;')



import functools
import unittest
from unittest import mock

import pytest
from sqlalchemy import MetaData, Table, Column, Integer, String

from aiomysql import sa


meta = MetaData()
tbl = Table('sa_tbl2', meta,
            Column('id', Integer, nullable=False,
                   primary_key=True),
            Column('name', String(255)))


def check_prepared_transactions(func):
    @functools.wraps(func)
    async def wrapper(self):
        conn = await self.loop.run_until_complete(self._connect())
        val = await conn.scalar('show max_prepared_transactions')
        if not val:
            raise unittest.SkipTest('Twophase transacions are not supported. '
                                    'Set max_prepared_transactions to '
                                    'a nonzero value')
        return func(self)
    return wrapper


async def start(conn):
    await conn.execute("DROP TABLE IF EXISTS sa_tbl2")
    await conn.execute("CREATE TABLE sa_tbl2 "
                       "(id serial, name varchar(255))")
    await conn.execute("INSERT INTO sa_tbl2 (name)"
                       "VALUES ('first')")
    await conn._connection.commit()


@pytest.fixture()
def sa_connect(connection, connection_creator):
    async def _connect(**kwargs):
        conn = await connection_creator(**kwargs)
        # TODO: fix this, should autocommit be enabled by default?
        await conn.autocommit(True)
        engine = mock.Mock()
        engine.dialect = sa.engine._dialect

        def release(*args):
            return
        engine.release = release

        ret = sa.SAConnection(conn, engine)
        return ret
    return _connect


@pytest.mark.run_loop
async def test_without_transactions(sa_connect):
    conn1 = await sa_connect()
    await start(conn1)

    conn2 = await sa_connect()
    res1 = await conn1.scalar(tbl.count())
    assert 1 == res1

    await conn2.execute(tbl.delete())

    res2 = await conn1.scalar(tbl.count())
    assert 0 == res2
    await conn1.close()
    await conn2.close()


@pytest.mark.run_loop
async def test_connection_attr(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr = await conn.begin()
    assert tr.connection is conn
    await conn.close()


@pytest.mark.run_loop
async def test_root_transaction(sa_connect):
    conn1 = await sa_connect()
    await start(conn1)
    conn2 = await sa_connect()

    tr = await conn1.begin()
    assert tr.is_active
    await conn1.execute(tbl.delete())

    res1 = await conn2.scalar(tbl.count())
    assert 1 == res1

    await tr.commit()

    assert not tr.is_active
    assert not conn1.in_transaction
    res2 = await conn2.scalar(tbl.count())
    assert 0 == res2
    await conn1.close()
    await conn2.close()


@pytest.mark.run_loop
async def test_root_transaction_rollback(sa_connect):
    conn1 = await sa_connect()
    await start(conn1)
    conn2 = await sa_connect()

    tr = await conn1.begin()
    assert tr.is_active
    await conn1.execute(tbl.delete())

    res1 = await conn2.scalar(tbl.count())
    assert 1 == res1

    await tr.rollback()

    assert not tr.is_active
    res2 = await conn2.scalar(tbl.count())
    assert 1 == res2
    await conn1.close()
    await conn2.close()


@pytest.mark.run_loop
async def test_root_transaction_close(sa_connect):
    conn1 = await sa_connect()
    await start(conn1)
    conn2 = await sa_connect()

    tr = await conn1.begin()
    assert tr.is_active
    await conn1.execute(tbl.delete())

    res1 = await conn2.scalar(tbl.count())
    assert 1 == res1

    await tr.close()

    assert not tr.is_active
    res2 = await conn2.scalar(tbl.count())
    assert 1 == res2
    await conn1.close()
    await conn2.close()


@pytest.mark.run_loop
async def test_rollback_on_connection_close(sa_connect):
    conn1 = await sa_connect()
    await start(conn1)
    conn2 = await sa_connect()

    tr = await conn1.begin()
    await conn1.execute(tbl.delete())

    res1 = await conn2.scalar(tbl.count())
    assert 1 == res1

    await conn1.close()

    res2 = await conn2.scalar(tbl.count())
    assert 1 == res2
    del tr
    await conn1.close()
    await conn2.close()


@pytest.mark.run_loop
async def test_root_transaction_commit_inactive(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr = await conn.begin()
    assert tr.is_active
    await tr.commit()
    assert not tr.is_active
    with pytest.raises(sa.InvalidRequestError):
        await tr.commit()
    await conn.close()


@pytest.mark.run_loop
async def test_root_transaction_rollback_inactive(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr = await conn.begin()
    assert tr.is_active
    await tr.rollback()
    assert not tr.is_active
    await tr.rollback()
    assert not tr.is_active
    await conn.close()


@pytest.mark.run_loop
async def test_root_transaction_double_close(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr = await conn.begin()
    assert tr.is_active
    await tr.close()
    assert not tr.is_active
    await tr.close()
    assert not tr.is_active
    await conn.close()


@pytest.mark.run_loop
async def test_inner_transaction_commit(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr1 = await conn.begin()
    tr2 = await conn.begin()
    assert tr2.is_active

    await tr2.commit()
    assert not tr2.is_active
    assert tr1.is_active

    await tr1.commit()
    assert not tr2.is_active
    assert not tr1.is_active
    await conn.close()


@pytest.mark.run_loop
async def test_inner_transaction_rollback(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr1 = await conn.begin()
    tr2 = await conn.begin()
    assert tr2.is_active
    await conn.execute(tbl.insert().values(name='aaaa'))

    await tr2.rollback()
    assert not tr2.is_active
    assert not tr1.is_active

    res = await conn.scalar(tbl.count())
    assert 1 == res
    await conn.close()


@pytest.mark.run_loop
async def test_inner_transaction_close(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr1 = await conn.begin()
    tr2 = await conn.begin()
    assert tr2.is_active
    await conn.execute(tbl.insert().values(name='aaaa'))

    await tr2.close()
    assert not tr2.is_active
    assert tr1.is_active
    await tr1.commit()

    res = await conn.scalar(tbl.count())
    assert 2 == res
    await conn.close()


@pytest.mark.run_loop
async def test_nested_transaction_commit(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr1 = await conn.begin_nested()
    tr2 = await conn.begin_nested()
    assert tr1.is_active
    assert tr2.is_active

    await conn.execute(tbl.insert().values(name='aaaa'))
    await tr2.commit()
    assert not tr2.is_active
    assert tr1.is_active

    res = await conn.scalar(tbl.count())
    assert 2 == res

    await tr1.commit()
    assert not tr2.is_active
    assert not tr1.is_active

    res = await conn.scalar(tbl.count())
    assert 2 == res
    await conn.close()


@pytest.mark.run_loop
async def test_nested_transaction_commit_twice(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr1 = await conn.begin_nested()
    tr2 = await conn.begin_nested()

    await conn.execute(tbl.insert().values(name='aaaa'))
    await tr2.commit()
    assert not tr2.is_active
    assert tr1.is_active

    await tr2.commit()
    assert not tr2.is_active
    assert tr1.is_active

    res = await conn.scalar(tbl.count())
    assert 2 == res

    await tr1.close()
    await conn.close()


@pytest.mark.run_loop
async def test_nested_transaction_rollback(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr1 = await conn.begin_nested()
    tr2 = await conn.begin_nested()
    assert tr1.is_active
    assert tr2.is_active

    await conn.execute(tbl.insert().values(name='aaaa'))
    await tr2.rollback()
    assert not tr2.is_active
    assert tr1.is_active

    res = await conn.scalar(tbl.count())
    assert 1 == res

    await tr1.commit()
    assert not tr2.is_active
    assert not tr1.is_active

    res = await conn.scalar(tbl.count())
    assert 1 == res
    await conn.close()


@pytest.mark.run_loop
async def test_nested_transaction_rollback_twice(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr1 = await conn.begin_nested()
    tr2 = await conn.begin_nested()

    await conn.execute(tbl.insert().values(name='aaaa'))
    await tr2.rollback()
    assert not tr2.is_active
    assert tr1.is_active

    await tr2.rollback()
    assert not tr2.is_active
    assert tr1.is_active

    await tr1.commit()
    res = await conn.scalar(tbl.count())
    assert 1 == res
    await conn.close()


@pytest.mark.run_loop
async def test_twophase_transaction_commit(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr = await conn.begin_twophase('sa_twophase')
    assert tr.xid == 'sa_twophase'
    await conn.execute(tbl.insert().values(name='aaaa'))

    await tr.prepare()
    assert tr.is_active

    await tr.commit()
    assert not tr.is_active

    res = await conn.scalar(tbl.count())
    assert 2 == res
    await conn.close()


@pytest.mark.run_loop
async def test_twophase_transaction_twice(sa_connect):
    conn = await sa_connect()
    await start(conn)
    tr = await conn.begin_twophase()
    with pytest.raises(sa.InvalidRequestError):
        await conn.begin_twophase()

    assert tr.is_active
    await tr.prepare()
    await tr.commit()
    await conn.close()


@pytest.mark.run_loop
async def test_transactions_sequence(sa_connect):
    conn = await sa_connect()
    await start(conn)

    await conn.execute(tbl.delete())

    assert conn._transaction is None

    tr1 = await conn.begin()
    assert tr1 is conn._transaction
    await conn.execute(tbl.insert().values(name='a'))
    res1 = await conn.scalar(tbl.count())
    assert 1 == res1

    await tr1.commit()
    assert conn._transaction is None

    tr2 = await conn.begin()
    assert tr2 is conn._transaction
    await conn.execute(tbl.insert().values(name='b'))
    res2 = await conn.scalar(tbl.count())
    assert 2 == res2
    await tr2.rollback()
    assert conn._transaction is None

    tr3 = await conn.begin()
    assert tr3 is conn._transaction
    await conn.execute(tbl.insert().values(name='b'))
    res3 = await conn.scalar(tbl.count())
    assert 2 == res3
    await tr3.commit()
    assert conn._transaction is None
    await conn.close()
